from __future__ import annotations
from functools import cached_property
from pathlib import Path
from datetime import datetime, timedelta
from typing import Any, Sequence

import numpy as np
import pandas as pd
from mikecore.DfsFactory import DfsBuilder, DfsFactory
from mikecore.DfsFile import DfsSimpleType, StatType, TimeAxisType
from mikecore.DfsFileFactory import DfsFileFactory
from mikecore.eum import eumQuantity

from .. import __dfs_version__
from ..dataset import Dataset, DataArray
from ._dfs import _get_item_info, _valid_item_numbers, _valid_timesteps
from ..eum import EUMType, EUMUnit, ItemInfo, TimeStepUnit


def _write_dfs0(
    filename: str | Path,
    dataset: Dataset,
    title: str = "",
    dtype: DfsSimpleType = DfsSimpleType.Float,
) -> None:
    filename = str(filename)

    factory = DfsFactory()
    builder = DfsBuilder.Create(title, "mikeio", __dfs_version__)
    builder.SetDataType(1)
    builder.SetGeographicalProjection(factory.CreateProjectionUndefined())

    system_start_time = dataset.time[0]

    if dataset.is_equidistant:
        if len(dataset.time) == 1:
            dt = 1.0  # TODO
        else:
            dt = (dataset.time[1] - dataset.time[0]).total_seconds()

        temporal_axis = factory.CreateTemporalEqCalendarAxis(
            TimeStepUnit.SECOND, system_start_time, 0, dt
        )
    else:
        temporal_axis = factory.CreateTemporalNonEqCalendarAxis(
            TimeStepUnit.SECOND, system_start_time
        )

    builder.SetTemporalAxis(temporal_axis)
    builder.SetItemStatisticsType(StatType.RegularStat)

    dfs_dtype = Dfs0._to_dfs_datatype(dtype)

    for da in dataset:
        newitem = builder.CreateDynamicItemBuilder()
        quantity = eumQuantity.Create(da.type, da.unit)
        newitem.Set(da.name, quantity, dfs_dtype)
        newitem.SetValueType(da.item.data_value_type)
        newitem.SetAxis(factory.CreateAxisEqD0())
        builder.AddDynamicItem(newitem.GetDynamicItemInfo())

    builder.CreateFile(filename)

    dfs = builder.GetFile()

    delete_value = dfs.FileInfo.DeleteValueFloat

    t_seconds = (dataset.time - dataset.time[0]).total_seconds().values

    data = np.array(dataset.to_numpy(), order="F").astype(np.float64).T
    data[np.isnan(data)] = delete_value

    if data.ndim == 2:
        data_to_write = np.concatenate([t_seconds.reshape(-1, 1), data], axis=1)
    else:
        data_to_write = np.concatenate(
            [np.atleast_2d(t_seconds), np.atleast_2d(data)], axis=1
        )
    dfs.WriteDfs0DataDouble(data_to_write)

    dfs.Close()


class Dfs0:
    """Class for reading/writing dfs0 files."""

    def __init__(self, filename: str | Path):
        """Create a Dfs0 object for reading, writing.

        Parameters
        ----------
        filename: str or Path
            File name including full path to the dfs0 file.

        """
        self._filename = str(filename)

        path = Path(filename)
        if not path.exists():
            raise FileNotFoundError(path)

        dfs = DfsFileFactory.DfsGenericOpen(str(path))
        self._source = dfs

        # Read items
        self._n_items = len(dfs.ItemInfo)
        self._items = _get_item_info(dfs.ItemInfo, list(range(self._n_items)))

        self._timeaxistype = dfs.FileInfo.TimeAxis.TimeAxisType

        if self._timeaxistype in {
            TimeAxisType.CalendarEquidistant,
            TimeAxisType.CalendarNonEquidistant,
        }:
            self._start_time: datetime = dfs.FileInfo.TimeAxis.StartDateTime
        else:  # relative time axis
            self._start_time = datetime(1970, 1, 1)

        # time
        self._n_timesteps: int = dfs.FileInfo.TimeAxis.NumberOfTimeSteps

        dfs.Close()

    def __repr__(self) -> str:
        out = ["<mikeio.Dfs0>"]
        out.append(f"timeaxis: {repr(self._timeaxistype)}")

        if self._n_items < 10:
            out.append("items:")
            for i, item in enumerate(self.items):
                out.append(f"  {i}:  {item}")
        else:
            out.append(f"number of items: {self._n_items}")

        return str.join("\n", out)

    def read(
        self,
        items: str | int | Sequence[str | int] | None = None,
        time: int | str | slice | None = None,
        **kwargs: Any,
    ) -> Dataset:
        """Read data from a dfs0 file.

        Parameters
        ----------
        items: list[int] or list[str], optional
            Read only selected items, by number (0-based), or by name
        time: int, str, datetime, pd.TimeStamp, sequence, slice or pd.DatetimeIndex, optional
            Read only selected time steps, by default None (=all)
        **kwargs: Any
            Additional keyword arguments are ignored

        Returns
        -------
        Dataset
            A Dataset with data dimensions [t]

        """
        path = Path(self._filename)
        if not path.exists():
            raise FileNotFoundError(f"File {path} not found")

        # read data from file
        fdata, ftime = self._read(self._filename)
        dfs = self._dfs

        # select items
        self._n_items = len(dfs.ItemInfo)
        item_numbers = _valid_item_numbers(dfs.ItemInfo, items)
        if items is not None:
            fdata = [fdata[it] for it in item_numbers]
            fitems = [self.items[it] for it in item_numbers]
        else:
            fitems = self.items
        ds = Dataset(fdata, ftime, fitems, validate=False)

        # select time steps
        self._n_timesteps = dfs.FileInfo.TimeAxis.NumberOfTimeSteps
        if self._timeaxistype == TimeAxisType.CalendarNonEquidistant and isinstance(
            time, str
        ):
            sel_time_step_str = time
            time_steps = None
        else:
            sel_time_step_str = None
            time_steps = None
            if time is not None:
                _, time_steps = _valid_timesteps(dfs.FileInfo, time)

        if time_steps:
            ds = ds.isel(time=time_steps)

        if sel_time_step_str:
            parts = sel_time_step_str.split(",")
            if len(parts) == 1:
                parts.append(parts[0])  # end=start

            if parts[0] == "":
                sel = slice(parts[1])  # stop only
            elif parts[1] == "":
                sel = slice(parts[0], None)  # start only
            else:
                sel = slice(parts[0], parts[1])
            ds = ds[sel]  # type: ignore

        return ds

    def _read(self, filename: str) -> tuple[list[np.ndarray], pd.DatetimeIndex]:
        """Read all data from a dfs0 file."""
        self._dfs = DfsFileFactory.DfsGenericOpen(filename)
        raw_data = self._dfs.ReadDfs0DataDouble()  # Bulk read the data

        self._dfs.Close()

        matrix = raw_data[:, 1:]
        # matrix[matrix == self._deletevalue] = np.nan
        matrix[matrix == self._dfs.FileInfo.DeleteValueDouble] = np.nan  # cutil
        matrix[matrix == self._dfs.FileInfo.DeleteValueFloat] = np.nan  # linux
        data = []
        for i in range(matrix.shape[1]):
            data.append(matrix[:, i])

        t_seconds = raw_data[:, 0]
        time = pd.to_datetime(t_seconds, unit="s", origin=self.start_time)
        time = time.round(freq="ms")  # accept nothing finer than milliseconds

        return data, time

    @staticmethod
    def _to_dfs_datatype(dtype: Any = None) -> DfsSimpleType:
        if dtype is None:
            return DfsSimpleType.Float

        if dtype in {np.float64, DfsSimpleType.Double, "double"}:
            return DfsSimpleType.Double

        if dtype in {np.float32, DfsSimpleType.Float, "float", "single"}:
            return DfsSimpleType.Float

        raise TypeError("Dfs files only support float or double")

    def to_dataframe(
        self, unit_in_name: bool = False, round_time: str = "ms"
    ) -> pd.DataFrame:
        """Read data from the dfs0 file and return a Pandas DataFrame.

        Parameters
        ----------
        unit_in_name: bool, optional
            include unit in column name, default False
        round_time: string, bool, optional
            round time to avoid problem with floating point inaccurcy, set to False to avoid rounding
        Returns
        -------
        pd.DataFrame

        """
        data, time = self._read(self._filename)
        items = self.items
        if unit_in_name:
            cols = [f"{item.name} ({item.unit.name})" for item in items]
        else:
            cols = [f"{item.name}" for item in items]
        df = pd.DataFrame(np.atleast_2d(data).T, index=time, columns=cols)

        if round_time:
            rounded_idx = pd.DatetimeIndex(time).round(round_time)
            df.index = pd.DatetimeIndex(rounded_idx, freq="infer")
        else:
            df.index = pd.DatetimeIndex(time, freq="infer")

        return df

    @staticmethod
    def from_dataframe(
        df: pd.DataFrame,
        filename: str,
        itemtype: EUMType | None = None,
        unit: EUMUnit | None = None,
        items: Sequence[ItemInfo] | None = None,
    ) -> None:
        """Create a dfs0 from a pandas Dataframe.

        Parameters
        ----------

        df: pd.DataFrame
            Dataframe with data
        filename: str
            filename to write output
        itemtype: EUMType, optional
            Same type for all items
        unit: EUMUnit, optional
            Same unit for all items
        items: list[ItemInfo]
            Different types, units for each items

        """
        return dataframe_to_dfs0(df, filename, itemtype, unit, items)

    @property
    def n_items(self) -> int:
        """Number of items."""
        return self._n_items

    @property
    def items(self) -> list[ItemInfo]:
        """List of items."""
        return self._items

    @property
    def start_time(self) -> datetime:
        """File start time."""
        return self._start_time

    @cached_property
    def end_time(self) -> datetime:
        if self._source.FileInfo.TimeAxis.IsEquidistant():
            dt = self._source.FileInfo.TimeAxis.TimeStep
            n_steps = self._source.FileInfo.TimeAxis.NumberOfTimeSteps
            timespan = dt * (n_steps - 1)
        else:
            timespan = self._source.FileInfo.TimeAxis.TimeSpan

        return self.start_time + timedelta(seconds=timespan)

    @property
    def n_timesteps(self) -> int:
        """Number of time steps."""
        return self._n_timesteps

    @property
    def timestep(self) -> float:
        """Time step size in seconds."""
        if self._timeaxistype == TimeAxisType.CalendarEquidistant:
            return self._source.FileInfo.TimeAxis.TimeStep  # type: ignore
        else:
            raise ValueError("Time axis type not supported")

    @property
    def time(self) -> pd.DatetimeIndex:
        """File all datetimes."""
        if self._timeaxistype == TimeAxisType.CalendarEquidistant:
            freq = pd.Timedelta(seconds=self.timestep)
            return pd.date_range(
                start=self.start_time,
                periods=self.n_timesteps,
                freq=freq,
            )

        elif self._timeaxistype == TimeAxisType.CalendarNonEquidistant:
            return self.read().time
        else:
            raise ValueError("Time axis type not supported")


def series_to_dfs0(
    self: pd.Series,
    filename: str,
    itemtype: EUMType | None = None,
    unit: EUMUnit | None = None,
    items: Sequence[ItemInfo] | None = None,
    title: str | None = None,
    dtype: Any | None = None,
) -> None:
    df = pd.DataFrame(self)
    df.to_dfs0(filename, itemtype, unit, items, title, dtype)


def dataframe_to_dfs0(
    self: pd.DataFrame,
    filename: str,
    itemtype: EUMType | None = None,
    unit: EUMUnit | None = None,
    items: Sequence[ItemInfo] | None = None,
    title: str = "",
    dtype: Any | None = None,
) -> None:
    """Create a dfs0.

    Parameters
    ----------
    self: pd.DataFrame
        Dataframe with data
    filename: str
        filename to write output
    itemtype: EUMType, optional
        Same type for all items
    unit: EUMUnit, optional
        Same unit for all items
    items: list[ItemInfo]
        Different types, units for each items
    title: str, optional
        Title of dfs0 file
    dtype : np.dtype, optional
            default np.float32

    """
    if not isinstance(self.index, pd.DatetimeIndex):
        raise ValueError(
            "Dataframe index must be a DatetimeIndex. Hint: pd.read_csv(..., parse_dates=True)"
        )

    ncol = self.values.shape[1]
    data = [self.values[:, i] for i in range(ncol)]

    if items is None:
        if itemtype is None:
            items = [ItemInfo(name) for name in self.columns]
        else:
            if unit is None:
                items = [ItemInfo(name, itemtype) for name in self.columns]
            else:
                items = [ItemInfo(name, itemtype, unit) for name in self.columns]

    das = {
        item.name: DataArray(data=d, item=item, time=self.index)
        for d, item in zip(data, items)
    }
    ds = Dataset(das)
    _write_dfs0(filename=filename, dataset=ds, title=title, dtype=dtype)


# Monkey patching onto Pandas classes
pd.DataFrame.to_dfs0 = dataframe_to_dfs0

pd.Series.to_dfs0 = series_to_dfs0



from __future__ import annotations
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from typing import Any, Literal, Sequence

import numpy as np
import pandas as pd
from mikecore.DfsFile import TimeAxisType
from mikecore.DfsFactory import DfsFactory
from mikecore.DfsuBuilder import DfsuBuilder
from mikecore.DfsuFile import DfsuFile, DfsuFileType
from mikecore.DfsFileFactory import DfsFileFactory
from mikecore.eum import eumQuantity, eumUnit
from tqdm import trange

from mikeio.spatial._utils import xy_to_bbox

from .. import __dfs_version__
from ..dataset import Dataset
from ..dfs._dfs import (
    _get_item_info,
    _read_item_time_step,
    _valid_item_numbers,
    _valid_timesteps,
)
from ..spatial import (
    GeometryFM2D,
)
from ..spatial import Grid2D
from .._track import _extract_track
from ._common import get_elements_from_source, get_nodes_from_source
from ..eum import ItemInfo, TimeStepUnit


def write_dfsu(filename: str | Path, data: Dataset) -> None:
    """Write a dfsu file.

    Parameters
    ----------
    filename: str
        dfsu filename
    data: Dataset
        Dataset to be written

    """
    filename = str(filename)

    geometry = data.geometry
    dfsu_filetype = DfsuFileType.Dfsu2D

    if geometry.is_layered:
        dfsu_filetype = geometry._type.value

    xn = geometry.node_coordinates[:, 0]
    yn = geometry.node_coordinates[:, 1]
    zn = geometry.node_coordinates[:, 2]

    elem_table = [np.array(e) + 1 for e in geometry.element_table]

    builder = DfsuBuilder.Create(dfsu_filetype)
    if dfsu_filetype != DfsuFileType.Dfsu2D:
        builder.SetNumberOfSigmaLayers(geometry.n_sigma_layers)

    builder.SetNodes(xn, yn, zn, geometry.codes)
    builder.SetElements(elem_table)

    factory = DfsFactory()
    proj = factory.CreateProjection(geometry.projection_string)
    builder.SetProjection(proj)

    if data.is_equidistant:
        temporal_axis = factory.CreateTemporalEqCalendarAxis(
            TimeStepUnit.SECOND, data.time[0], 0, data.timestep
        )
    else:
        temporal_axis = factory.CreateTemporalNonEqCalendarAxis(
            TimeStepUnit.SECOND, data.time[0]
        )
    builder.SetTemporalAxis(temporal_axis)
    builder.SetZUnit(eumUnit.eumUmeter)

    if dfsu_filetype != DfsuFileType.Dfsu2D:
        builder.SetNumberOfSigmaLayers(geometry.n_sigma_layers)

    for item in data.items:
        builder.AddDynamicItem(item.name, eumQuantity.Create(item.type, item.unit))

    builder.ApplicationTitle = "mikeio"
    builder.ApplicationVersion = __dfs_version__
    dfs = builder.CreateFile(filename)

    write_dfsu_data(dfs, data, geometry.is_layered)


def write_dfsu_data(dfs: DfsuFile, ds: Dataset, is_layered: bool) -> None:
    n_time_steps = len(ds.time)
    data = ds

    if data.is_equidistant:
        t_rel = np.zeros(data.n_timesteps)
    else:
        t_rel = (data.time - data.time[0]).total_seconds()

    for i in range(n_time_steps):
        if is_layered:
            if "time" in data.dims:
                assert data._zn is not None
                zn = data._zn[i]
            else:
                zn = data._zn
            dfs.WriteItemTimeStepNext(t_rel[i], zn.astype(np.float32))
        for da in data:
            if "time" in data.dims:
                d = da.to_numpy()[i, :]
            else:
                d = da.to_numpy()
            d[np.isnan(d)] = data.deletevalue
            dfs.WriteItemTimeStepNext(t_rel[i], d.astype(np.float32))
    dfs.Close()


def _validate_elements_and_geometry_sel(elements: Any, **kwargs: Any) -> None:
    """Check that only one of elements, area, x, y is selected."""
    used_kwargs = [key for key, val in kwargs.items() if val is not None]

    if elements is not None and len(used_kwargs) > 0:
        raise ValueError(f"Cannot select both {used_kwargs} and elements!")

    if "area" in used_kwargs and ("x" in used_kwargs or "y" in used_kwargs):
        raise ValueError("Cannot select both x,y and area!")


@dataclass
class _DfsuInfo:
    filename: str
    type: DfsuFileType
    start_time: datetime
    equidistant: bool
    timestep: float
    n_timesteps: int
    items: list[ItemInfo]
    deletevalue: float


def _get_dfsu_info(filename: str | Path) -> _DfsuInfo:
    filename = str(filename)
    path = Path(filename)
    if not path.exists():
        raise FileNotFoundError(f"file {path} does not exist!")
    dfs = DfsuFile.Open(filename)
    type = DfsuFileType(dfs.DfsuFileType)
    deletevalue = dfs.DeleteValueFloat

    timestep = dfs.TimeStepInSeconds
    items = _get_item_info(dfs.ItemInfo)
    equidistant = dfs.FileInfo.TimeAxis.TimeAxisType == TimeAxisType.CalendarEquidistant
    dfs.Close()
    return _DfsuInfo(
        filename=filename,
        type=type,
        timestep=timestep,
        equidistant=equidistant,
        n_timesteps=dfs.NumberOfTimeSteps,
        items=items,
        start_time=dfs.FileInfo.TimeAxis.StartDateTime,
        deletevalue=deletevalue,
    )


# class _Dfsu:
#     show_progress = False

#     def __init__(self, filename: str | Path) -> None:
#         """
#         Create a Dfsu object

#         Parameters
#         ---------
#         filename: str
#             dfsu filename
#         """
#         info = _get_dfsu_info(filename)
#         self._filename = info.filename
#         self._type = info.type
#         self._deletevalue = info.deletevalue
#         self._time = info.time
#         self._timestep = info.timestep
#         self._items = info.items

#     def __repr__(self):
#         out = [f"<mikeio.{self.__class__.__name__}>"]

#         if self._type is not DfsuFileType.DfsuSpectral0D:
#             if self._type is not DfsuFileType.DfsuSpectral1D:
#                 out.append(f"number of elements: {self.geometry.n_elements}")
#             out.append(f"number of nodes: {self.geometry.n_nodes}")
#         if self.geometry.is_spectral:
#             if self.geometry.n_directions > 0:
#                 out.append(f"number of directions: {self.geometry.n_directions}")
#             if self.geometry.n_frequencies > 0:
#                 out.append(f"number of frequencies: {self.geometry.n_frequencies}")
#         if self.geometry.projection_string:
#             out.append(f"projection: {self.geometry.projection_string}")
#         if self.geometry.is_layered:
#             out.append(f"number of sigma layers: {self.geometry.n_sigma_layers}")
#         if (
#             self._type == DfsuFileType.DfsuVerticalProfileSigmaZ
#             or self._type == DfsuFileType.Dfsu3DSigmaZ
#         ):
#             out.append(
#                 f"max number of z layers: {self.geometry.n_layers - self.geometry.n_sigma_layers}"
#             )
#         if self.n_items < 10:
#             out.append("items:")
#             for i, item in enumerate(self.items):
#                 out.append(f"  {i}:  {item}")
#         else:
#             out.append(f"number of items: {self.geometry.n_items}")
#         if self.n_timesteps == 1:
#             out.append(f"time: time-invariant file (1 step) at {self.time[0]}")
#         else:
#             out.append(
#                 f"time: {str(self.time[0])} - {str(self.time[-1])} ({self.n_timesteps} records)"
#             )
#         return str.join("\n", out)

#     def _read_items(self, filename: str) -> list[ItemInfo]:
#         dfs = DfsuFile.Open(filename)
#         items = _get_item_info(dfs.ItemInfo)
#         dfs.Close()
#         return items

#     @property
#     def geometry(self):
#         return self._geometry

#     @property
#     def deletevalue(self) -> float:
#         """File delete value"""
#         return self._deletevalue

#     @property
#     def n_items(self) -> int:
#         """Number of items"""
#         return len(self.items)

#     @property
#     def items(self) -> list[ItemInfo]:
#         """List of items"""
#         return self._items

#     @property
#     def start_time(self) -> pd.Timestamp:
#         """File start time"""
#         return self._time[0]

#     @property
#     def n_timesteps(self) -> int:
#         """Number of time steps"""
#         return len(self._time)

#     @property
#     def timestep(self) -> float:
#         """Time step size in seconds"""
#         return self._timestep

#     @property
#     def end_time(self) -> pd.Timestamp:
#         """File end time"""
#         return self._time[-1]

#     @property
#     def time(self) -> pd.DatetimeIndex:
#        return self._time


class Dfsu2DH:
    """Class for reading/writing dfsu 2d horizontal files."""

    show_progress = False

    def __init__(self, filename: str | Path) -> None:
        info = _get_dfsu_info(filename)
        self._filename = info.filename
        self._type = info.type
        self._deletevalue = info.deletevalue
        self._equidistant = info.equidistant
        self._start_time = info.start_time
        self._timestep = info.timestep
        self._n_timesteps = info.n_timesteps
        self._items = info.items
        self._geometry = self._read_geometry(self._filename)

    def __repr__(self) -> str:
        out = [f"<mikeio.{self.__class__.__name__}>"]

        out.append(f"number of elements: {self.geometry.n_elements}")
        out.append(f"number of nodes: {self.geometry.n_nodes}")
        out.append(f"projection: {self.geometry.projection_string}")
        if self.n_items < 10:
            out.append("items:")
            for i, item in enumerate(self.items):
                out.append(f"  {i}:  {item}")
        else:
            out.append(f"number of items: {self.n_items}")
        if self.n_timesteps == 1:
            out.append(f"time: time-invariant file (1 step) at {self.time[0]}")
        else:
            out.append(
                f"time: {str(self.time[0])} - {str(self.time[-1])} ({self.n_timesteps} records)"
            )
        return str.join("\n", out)

    @property
    def geometry(self) -> Any:
        return self._geometry

    @property
    def deletevalue(self) -> float:
        """File delete value."""
        return self._deletevalue

    @property
    def n_items(self) -> int:
        """Number of items."""
        return len(self.items)

    @property
    def items(self) -> list[ItemInfo]:
        """List of items."""
        return self._items

    @property
    def start_time(self) -> datetime:
        """File start time."""
        return self._start_time

    @property
    def n_timesteps(self) -> int:
        """Number of time steps."""
        return self._n_timesteps

    @property
    def timestep(self) -> float:
        """Time step size in seconds."""
        return self._timestep

    @property
    def end_time(self) -> pd.Timestamp:
        """File end time."""
        if self._equidistant:
            return self.time[-1]
        else:
            # read the last timestep
            ds = self.read(items=0, time=-1)
            return ds.time[-1]

    @property
    def time(self) -> pd.DatetimeIndex:
        if self._equidistant:
            return pd.date_range(
                start=self.start_time,
                periods=self.n_timesteps,
                freq=f"{int(self.timestep)}s",
            )
        else:
            raise NotImplementedError(
                "Non-equidistant time axis. Read the data to get time."
            )

    @staticmethod
    def _read_geometry(filename: str) -> GeometryFM2D:
        dfs = DfsuFile.Open(filename)
        dfsu_type = DfsuFileType(dfs.DfsuFileType)

        node_table = get_nodes_from_source(dfs)
        el_table = get_elements_from_source(dfs)

        geometry = GeometryFM2D(
            node_coordinates=node_table.coordinates,
            element_table=el_table.connectivity,
            codes=node_table.codes,
            projection=dfs.Projection.WKTString,
            dfsu_type=dfsu_type,
            element_ids=el_table.ids,
            node_ids=node_table.ids,
            validate=False,
        )
        dfs.Close()
        return geometry

    def read(
        self,
        *,
        items: str | int | Sequence[str | int] | None = None,
        time: int | str | slice | None = None,
        elements: Sequence[int] | np.ndarray | None = None,
        area: tuple[float, float, float, float] | None = None,
        x: float | Sequence[float] | None = None,
        y: float | Sequence[float] | None = None,
        keepdims: bool = False,
        dtype: Any = np.float32,
        error_bad_data: bool = True,
        fill_bad_data_value: float = np.nan,
    ) -> Dataset:
        """Read data from a dfsu file.

        Parameters
        ---------
        items: list[int] or list[str], optional
            Read only selected items, by number (0-based), or by name
        time: int, str, datetime, pd.TimeStamp, sequence, slice or pd.DatetimeIndex, optional
            Read only selected time steps, by default None (=all)
        keepdims: bool, optional
            When reading a single time step only, should the time-dimension be kept
            in the returned Dataset? by default: False
        area: list[float], optional
            Read only data inside (horizontal) area given as a
            bounding box (tuple with left, lower, right, upper)
            or as list of coordinates for a polygon, by default None
        x, y: float or list[float], optional
            Read only data for elements containing the (x,y) points(s),
            by default None
        elements: list[int], optional
            Read only selected element ids, by default None
        error_bad_data: bool, optional
            raise error if data is corrupt, by default True,
        fill_bad_data_value:
            fill value for to impute corrupt data, used in conjunction with error_bad_data=False
            default np.nan
        dtype: Any, optional
            Data type to read, by default np.float32

        Returns
        -------
        Dataset
            A Dataset with data dimensions [t,elements]

        """
        if dtype not in [np.float32, np.float64]:
            raise ValueError("Invalid data type. Choose np.float32 or np.float64")
        dfs = DfsuFile.Open(self._filename)

        single_time_selected, time_steps = _valid_timesteps(dfs, time)

        _validate_elements_and_geometry_sel(elements, area=area, x=x, y=y)
        if area is not None:
            elements = self.geometry._elements_in_area(area)

        if (x is not None) or (y is not None):
            elements = self.geometry.find_index(x=x, y=y)

        if elements is None:
            geometry = self.geometry
            n_elems = geometry.n_elements
        else:
            elements = [elements] if np.isscalar(elements) else list(elements)  # type: ignore
            n_elems = len(elements)
            geometry = self.geometry.elements_to_geometry(elements)

        item_numbers = _valid_item_numbers(dfs.ItemInfo, items)
        n_items = len(item_numbers)

        t_rel = np.zeros(len(time_steps))

        n_steps = len(time_steps)
        shape: tuple[int, ...] = (
            (n_elems,)
            if (single_time_selected and not keepdims)
            else (n_steps, n_elems)
        )
        data_list: list[np.ndarray] = [
            np.ndarray(shape=shape, dtype=dtype) for _ in range(n_items)
        ]

        for i in trange(n_steps, disable=not self.show_progress):
            for item in range(n_items):
                dfs, d, t_rel[i] = _read_item_time_step(
                    dfs=dfs,
                    filename=self._filename,
                    time=time,
                    item_numbers=item_numbers,
                    deletevalue=self.deletevalue,
                    shape=shape,
                    item=item,
                    it=time_steps[i],
                    error_bad_data=error_bad_data,
                    fill_bad_data_value=fill_bad_data_value,
                )

                if elements is not None:
                    d = d[elements]

                if single_time_selected and not keepdims:
                    data_list[item] = d
                else:
                    data_list[item][i] = d

        dfs.Close()

        dims: tuple[str, ...] = ("time", "element")

        if single_time_selected and not keepdims:
            dims = ("element",)

        if elements is not None and len(elements) == 1:
            # squeeze point data
            dims = tuple([d for d in dims if d != "element"])
            data_list = [np.squeeze(d, axis=-1) for d in data_list]

        time = pd.to_datetime(t_rel, unit="s", origin=self.start_time)
        item_infos = _get_item_info(dfs.ItemInfo, item_numbers)

        return Dataset(
            data_list,
            time,
            item_infos,
            geometry=geometry,
            dims=dims,
            validate=False,
            dt=self.timestep,
        )

    def append(self, ds: Dataset, validate: bool = True) -> None:
        """Append data to an existing dfsu file.

        Parameters
        ----------
        ds: Dataset
            Dataset to be appended
        validate: bool, optional
            Validate that the items and geometry match, by default True

        """
        if validate:
            if ds.geometry != self.geometry:
                raise ValueError("The geometry of the dataset to append does not match")

            for item_s, item_o in zip(ds.items, self.items):
                if item_s != item_o:
                    raise ValueError(
                        f"Item in dataset {item_s.name} does not match {item_o.name}"
                    )

        dfs = DfsFileFactory.DfsuFileOpenAppend(str(self._filename), parameters=None)
        write_dfsu_data(dfs=dfs, ds=ds, is_layered=False)
        info = _get_dfsu_info(self._filename)
        self._n_timesteps = info.n_timesteps

    def get_overset_grid(
        self,
        dx: float | None = None,
        dy: float | None = None,
        nx: int | None = None,
        ny: int | None = None,
        buffer: float = 0.0,
    ) -> Grid2D:
        """get a 2d grid that covers the domain by specifying spacing or shape.

        Parameters
        ----------
        dx : float, optional
            grid resolution in x-direction (or in x- and y-direction)
        dy : float, optional
            grid resolution in y-direction
        nx : int, optional
            number of points in x-direction,
            by default None (the value will be inferred)
        ny : int, optional
            number of points in y-direction,
            by default None (the value will be inferred)
        buffer : float, optional
            positive to make the area larger, default=0
            can be set to a small negative value to avoid NaN
            values all around the domain.

        Returns
        -------
        <mikeio.Grid2D>
            2d grid

        """
        nc = self.geometry.geometry2d.node_coordinates
        bbox = xy_to_bbox(nc, buffer=buffer)
        return Grid2D(
            bbox=bbox,
            dx=dx,
            dy=dy,
            nx=nx,
            ny=ny,
            projection=self.geometry.projection_string,
        )

    def _dfs_read_item_time_func(
        self, item: int, step: int
    ) -> tuple[np.ndarray, pd.Timestamp]:
        dfs = DfsuFile.Open(self._filename)
        itemdata = dfs.ReadItemTimeStep(item + 1, step)

        return itemdata.Data, itemdata.Time

    def extract_track(
        self,
        track: pd.DataFrame,
        items: int | str | Sequence[int | str] | None = None,
        method: Literal["nearest", "inverse_distance"] = "nearest",
        dtype: Any = np.float32,
    ) -> Dataset:
        """Extract track data from a dfsu file.

        Parameters
        ---------
        track: pandas.DataFrame
            with DatetimeIndex and (x, y) of track points as first two columns
            x,y coordinates must be in same coordinate system as dfsu
        track: str
            filename of csv or dfs0 file containing t,x,y
        items: list[int] or list[str], optional
            Extract only selected items, by number (0-based), or by name
        method: str, optional
            Spatial interpolation method ('nearest' or 'inverse_distance')
            default='nearest'
        dtype: Any, optional
            Data type to read, by default np.float32

        Returns
        -------
        Dataset
            A dataset with data dimension t
            The first two items will be x- and y- coordinates of track

        Examples
        --------
        ```{python}
        import mikeio

        ds = (
            mikeio.open("../data/NorthSea_HD_and_windspeed.dfsu")
                  .extract_track("../data/altimetry_NorthSea_20171027.csv")
            )
        ds
        ```

        """
        dfs = DfsuFile.Open(self._filename)

        item_numbers = _valid_item_numbers(dfs.ItemInfo, items)
        items = _get_item_info(dfs.ItemInfo, item_numbers)
        _, time_steps = _valid_timesteps(dfs, time_steps=None)

        res = _extract_track(
            deletevalue=self.deletevalue,
            start_time=self.start_time,
            end_time=self.end_time,
            timestep=self.timestep,
            geometry=self.geometry,
            n_elements=self.geometry.n_elements,
            track=track,
            items=deepcopy(items),
            time_steps=time_steps,
            item_numbers=item_numbers,
            method=method,
            dtype=dtype,
            data_read_func=self._dfs_read_item_time_func,
        )
        dfs.Close()
        return res



from abc import ABC, abstractmethod

from dataclasses import dataclass
from typing import Any, Sequence

from mikecore.Projections import MapProjection


@dataclass
class BoundingBox:
    """Bounding box for spatial data."""

    left: float
    bottom: float
    right: float
    top: float

    def __post_init__(self) -> None:
        if self.left > self.right:
            raise ValueError(
                f"Invalid x axis, left: {self.left} must be smaller than right: {self.right}"
            )

        if self.bottom > self.top:
            raise ValueError(
                f"Invalid y axis, bottom: {self.bottom} must be smaller than top: {self.top}"
            )

    def overlaps(self, other: "BoundingBox") -> bool:
        """Check if two bounding boxes overlap."""
        return not (
            self.left > other.right
            or self.bottom > other.top
            or self.right < other.left
            or self.top < other.bottom
        )

    @staticmethod
    def parse(
        values: "BoundingBox" | Sequence[float],
    ) -> "BoundingBox":
        match values:
            case BoundingBox():
                bbox = values
            case left, bottom, right, top:
                bbox = BoundingBox(left, bottom, right, top)
            case _:
                raise ValueError(
                    "values must be a bounding box of coordinates e.g. (-10.0, 10.0 20.0, 30.0)"
                )
        return bbox


class _Geometry(ABC):
    def __init__(self, projection: str = "LONG/LAT") -> None:
        if not MapProjection.IsValid(projection):
            raise ValueError(f"{projection=} is not a valid projection string")

        self._projstr = projection

    @property
    def projection_string(self) -> str:
        """The projection string."""
        return self._projstr

    @property
    def projection(self) -> str:
        """The projection."""
        return self._projstr

    @property
    def is_geo(self) -> bool:
        """Are coordinates geographical (LONG/LAT)?"""
        return self._projstr == "LONG/LAT"

    @property
    def is_local_coordinates(self) -> bool:
        """Are coordinates relative (NON-UTM)?"""
        return self._projstr == "NON-UTM"

    @property
    @abstractmethod
    def ndim(self) -> int:
        pass

    @property
    @abstractmethod
    def default_dims(self) -> tuple[str, ...]:
        pass


class GeometryUndefined(_Geometry):
    def __repr__(self) -> str:
        return "GeometryUndefined()"

    @property
    def ndim(self) -> int:
        raise NotImplementedError()

    @property
    def default_dims(self) -> tuple[str, ...]:
        raise NotImplementedError()


class GeometryPoint2D(_Geometry):
    def __init__(self, x: float, y: float, projection: str = "LONG/LAT"):
        super().__init__(projection)
        self.x = x
        self.y = y

    @property
    def default_dims(self) -> tuple[str, ...]:
        return ()

    def __repr__(self) -> str:
        return f"GeometryPoint2D(x={self.x}, y={self.y})"

    # TODO should we use wkt here
    # def __str__(self) -> str:
    #    return self.wkt

    @property
    def wkt(self) -> str:
        return f"POINT ({self.x} {self.y})"

    @property
    def ndim(self) -> int:
        """Geometry dimension."""
        return 0

    def to_shapely(self) -> Any:
        from shapely.geometry import Point

        return Point(self.x, self.y)


class GeometryPoint3D(_Geometry):
    def __init__(self, x: float, y: float, z: float, projection: str = "LONG/LAT"):
        super().__init__(projection)

        self.x = x
        self.y = y
        self.z = z

    def __repr__(self) -> str:
        return f"GeometryPoint3D(x={self.x}, y={self.y}, z={self.z})"

    @property
    def default_dims(self) -> tuple[str, ...]:
        return ()

    @property
    def wkt(self) -> str:
        return f"POINT Z ({self.x} {self.y} {self.z})"

    @property
    def ndim(self) -> int:
        return 0

    def to_shapely(self) -> Any:
        from shapely.geometry import Point

        return Point(self.x, self.y, self.z)



from __future__ import annotations
from typing import TYPE_CHECKING, overload
import numpy as np

if TYPE_CHECKING:
    from .dataset import Dataset, DataArray

from .spatial import GeometryUndefined


def get_idw_interpolant(distances: np.ndarray, p: float = 2) -> np.ndarray:
    """IDW interpolant for 2d array of distances.

    https://pro.arcgis.com/en/pro-app/help/analysis/geostatistical-analyst/how-inverse-distance-weighted-interpolation-works.htm

    Parameters
    ----------
    distances : array-like
        distances between interpolation point and grid point
    p : float, optional
        power of inverse distance weighting, default=2

    Returns
    -------
    np.array
        weights

    """
    is_1d = distances.ndim == 1
    if is_1d:
        distances = np.atleast_2d(distances)

    MIN_DISTANCE = 1e-8
    weights = np.zeros(distances.shape)

    match = distances[:, 0] < MIN_DISTANCE
    weights[match, 0] = 1

    weights[~match, :] = 1 / distances[~match, :] ** p
    denom = weights[~match, :].sum(axis=1).reshape(-1, 1)  # *np.ones((1,n_nearest))
    weights[~match, :] = weights[~match, :] / denom

    if is_1d:
        weights = weights[0]
    return weights


@overload
def interp2d(
    data: np.ndarray | DataArray,
    elem_ids: np.ndarray,
    weights: np.ndarray | None = None,
    shape: tuple[int, ...] | None = None,
) -> np.ndarray: ...


@overload
def interp2d(
    data: Dataset,
    elem_ids: np.ndarray,
    weights: np.ndarray | None = None,
    shape: tuple[int, ...] | None = None,
) -> Dataset: ...


def interp2d(
    data: Dataset | DataArray | np.ndarray,
    elem_ids: np.ndarray,
    weights: np.ndarray | None = None,
    shape: tuple[int, ...] | None = None,
) -> Dataset | np.ndarray:
    """interp spatially in data (2d only).

    Parameters
    ----------
    data : mikeio.Dataset, DataArray, or ndarray
        dfsu data
    elem_ids : ndarray(int)
        n sized array of 1 or more element ids used for interpolation
    weights : ndarray(float), optional
        weights with same size as elem_ids used for interpolation
    shape: tuple, optional
            reshape output

    Returns
    -------
    ndarray, Dataset, or DataArray
        spatially interped data with same type and shape as input

    Examples
    --------
    >>> elem_ids, weights = dfs.get_spatial_interpolant(coords)
    >>> dsi = interp2d(ds, elem_ids, weights)

    """
    from .dataset import DataArray, Dataset

    if isinstance(data, Dataset):
        ds = data.copy()

        ni = len(elem_ids)

        interp_data_vars = {}

        for da in ds:
            key = da.name
            if "time" not in da.dims:
                idatitem = _interp_itemstep(da.to_numpy(), elem_ids, weights)
                if shape:
                    idatitem = idatitem.reshape(*shape)

            else:
                nt, _ = da.shape
                # use dtype of da
                idatitem = np.empty(shape=(nt, ni), dtype=da.values.dtype)
                for step in range(nt):
                    idatitem[step, :] = _interp_itemstep(
                        da[step].to_numpy(), elem_ids, weights
                    )
                if shape:
                    idatitem = idatitem.reshape((nt, *shape))

            dims = ("time", "element")  # TODO is this the best?
            interp_data_vars[key] = DataArray(
                data=idatitem,
                time=da.time,
                dims=dims,
                item=da.item,
                geometry=GeometryUndefined(),
            )

        new_ds = Dataset(interp_data_vars, validate=False)
        return new_ds

    if isinstance(data, DataArray):
        # TODO why doesn't this return a DataArray?
        data = data.to_numpy()

    if isinstance(data, np.ndarray):
        if data.ndim == 1:
            # data is single item and single time step
            idatitem = _interp_itemstep(data, elem_ids, weights)
            if shape:
                idatitem = idatitem.reshape(*shape)
            return idatitem

    ni = len(elem_ids)
    datitem = data
    nt, _ = datitem.shape
    idatitem = np.empty(shape=(nt, ni), dtype=datitem.dtype)
    for step in range(nt):
        idatitem[step, :] = _interp_itemstep(datitem[step], elem_ids, weights)
    if shape:
        idatitem = idatitem.reshape((nt, *shape))
    return idatitem


def _interp_itemstep(
    data: np.ndarray,
    elem_ids: np.ndarray,
    weights: np.ndarray | None = None,
) -> np.ndarray:
    """Interpolate a single item and time step.

    Parameters
    ----------
    data : ndarray
        data to interpolate
    elem_ids : ndarray(int)
        n sized array of 1 or more element ids used for interpolation
    weights : ndarray(float), optional
        weights with same size as elem_ids used for interpolation

    Returns
    -------
    ndarray
        spatially interpolated data

    Notes
    -----
    This function is used internally by interp2d

    """
    if weights is None:
        return data[elem_ids]
    else:
        idat = data[elem_ids] * weights
        return np.sum(idat, axis=1) if weights.ndim == 2 else idat



"""Functionality related to the DHI EUM scientific type system.


Examples
--------
>>> mikeio.EUMType.Temperature
Temperature
>>> mikeio.EUMType.Temperature.value
100006
>>> mikeio.EUMType.Temperature.units
[degree Celsius, degree Fahrenheit, degree Kelvin]
>>> mikeio.EUMUnit.degree_Celsius
degree Celsius
>>> mikeio.EUMUnit.degree_Celsius.value
2800
>>>

"""

from __future__ import annotations
from enum import IntEnum
from typing import Any, Sequence, Literal

import pandas as pd
from mikecore.DfsFile import DataValueType, DfsDynamicItemInfo
from mikecore.eum import eumUnit, eumItem, eumWrapper

from ..exceptions import InvalidDataValueType


def _type_list(search: str | None = None) -> dict[eumItem, str]:
    """Get a dictionary of the EUM items.

    Notes
    -----
    An alternative to `type_list` is to use `mikeio.eum.Item`

    Parameters
    ----------
    search: str, optional
        a search string (caseinsensitive) to filter out results

    Returns
    -------
    dict
        names and codes for EUM items

    """
    items = {}
    check = True
    i = 1
    while check:
        d = eumWrapper.eumGetItemTypeSeq(i)
        if d[0] is True:
            items[d[1]] = d[2]
            i += 1
        else:
            check = False

    if search is not None:
        search = search.lower()
        items = dict(
            [
                [key, value]
                for key, value in items.items()
                if search in value.lower() or search == value.lower()
            ]
        )

    return items


def _unit_list(eum_type: int) -> dict[str, eumUnit]:
    """Get a dictionary of valid units.

    Parameters
    ----------
    eum_type: int
        EUM variable type, e.g. 100006 or EUMType.Temperature

    Returns
    -------
    dict
        names and codes for valid units

    """
    items = {}
    n_units_for_eum_type = eumWrapper.eumGetItemUnitCount(eum_type)
    for i in range(n_units_for_eum_type):
        _, value, key = eumWrapper.eumGetItemUnitSeq(eum_type, i + 1)
        items[key] = value

    return items


class TimeAxisType(IntEnum):
    EquidistantRelative = 1
    NonEquidistantRelative = 2
    EquidistantCalendar = 3
    NonEquidistantCalendar = 4

    def __repr__(self) -> str:
        return self.name


class TimeStepUnit(IntEnum):
    SECOND = 1400
    MINUTE = 1401
    HOUR = 1402
    DAY = 1403
    MONTH = 1405
    YEAR = 1404


class EUMType(IntEnum):
    """EUM type.

    Examples
    --------
    ```{python}
    import mikeio
    mikeio.EUMType.Temperature
    ```

    ```{python}
    mikeio.EUMType.Temperature.units
    ```

    """

    Water_Level = 100000
    Discharge = 100001
    Wind_Velocity = 100002
    Wind_Direction = 100003
    Rainfall = 100004
    Evaporation = 100005
    Temperature = 100006
    Concentration = 100007
    Bacteria_Concentration = 100008
    Resistance_factor = 100009
    Sediment_Transport = 100010
    Bottom_level = 100011
    Bottom_level_change = 100012
    Sediment_fraction = 100013
    Sediment_fraction_change = 100014
    Gate_Level = 100015
    Flow_velocity = 100016
    Density = 100017
    Dam_breach_level = 100018
    Dam_breach_width = 100019
    Dam_breach_slope = 100020
    Sunshine = 100021
    Sun_radiation = 100022
    Relative_humidity = 100023
    Salinity = 100024
    Surface_Slope = 100025
    Flow_Area = 100026
    Flow_Width = 100027
    Hydraulic_Radius = 100028
    Resistance_Radius = 100029
    Mannings_M = 100030
    Mannings_n = 100031
    Chezy_No = 100032
    Conveyance = 100033
    Froude_No = 100034
    Water_Volume = 100035
    Flooded_Area = 100036
    Water_Volume_Error = 100037
    Acc_Water_Volume_Error = 100038
    Component_Mass = 100039
    Component_Mass_Error = 100040
    Acc_Component_Mass_Error = 100041
    Relative_Component_Mass_Error = 100042
    Relative_Acc_Component_Mass_Error = 100043
    Component_Decay = 100044
    Acc_Component_Decay = 100045
    Component_Transport = 100046
    Acc_Component_Transport = 100047
    Component_Disp_Transport = 100048
    Acc_Component_Disp_Transport = 100049
    Component_Conv_Transport = 100050
    Acc_Component_Conv_Transport = 100051
    Acc_Sediment_transport = 100052
    Dune_length = 100053
    Dune_height = 100054
    Bed_sediment_load = 100055
    Suspended_sediment_load = 100056
    Irrigation = 100057
    Relative_moisture_content = 100058
    Ground_water_depth = 100059
    Snow_Water_Content = 100060
    Infiltration = 100061
    Recharge = 100062
    OF1_Flow = 100063
    IF1_Flow = 100064
    CapillaryFlux = 100065
    SurfStorage_OF1 = 100066
    SurfStorage_OF0 = 100067
    Sediment_layer = 100068
    Bed_level = 100069
    Rainfall_Intensity = 100070
    Production_rate = 100071
    Sediment_mass = 100072
    Primary_production = 100073
    Vol_specific_prod_rate = 100074
    Secchi_depth = 100075
    Acc_Sediment_Mass = 100076
    Sediment_Mass_per_m = 100077
    Surface_Elevation = 100078
    Bathymetry = 100079
    Flow_Flux = 100080
    Bed_sediment_load_per_m = 100081
    Suspended_load_per_m = 100082
    Sediment_transport_per_m = 100083
    Wave_height = 100084
    Wave_period = 100085
    Wave_frequency = 100086
    Potential_evaporation_rate = 100087
    Rainfall_rate = 100088
    Water_Flow = 100089
    Return_Flow_Fraction = 100090
    Linear_Routing_Coefficient = 100091
    Specific_runoff = 100092
    Machine_Efficiency = 100093
    Target_power = 100094
    Wave_direction = 100095
    Accumulated_transport_per_meter = 100096
    Significant_wave_height = 100097
    Critical_Shields_parameter = 100098
    AngleBedVelocity = 100099
    Profile_number = 100100
    Climate_number = 100101
    Spectral_description = 100102
    Spreading_factor = 100103
    Reference_point_number = 100104
    Wind_friction_factor = 100105
    Wave_Disturbance_Coefficient = 100106
    Time_of_first_wave_arrival = 100107
    Surface_Curvature = 100108
    Radiation_Stress = 100109
    Spectral_density = 100120
    Frequency_integrated_spectral_density = 100121
    Directional_integrated_spectral_density = 100122
    Viscosity = 100123
    Standard_deviation_DSD = 100124
    Beach_position = 100125
    Trench_position = 100126
    Grain_diameter = 100127
    Settling_velocity = 100128
    Geometrical_deviation = 100129
    Breaking_wave = 100130
    Dune_position = 100131
    Contour_angle = 100132
    Flow_direction = 100133
    Bed_slope = 100134
    Surface_area = 100135
    Catchment_area = 100136
    Roughness = 100137
    Active_Depth = 100138
    Sediment_Gradation = 100139
    Groundwater_recharge = 100140
    Solute_flux = 100141
    River_structure_geometry = 100142
    River_chainage = 100143
    Dimensionless_factor = 100144
    Dimensionless_exponent = 100145
    Storage_depth = 100146
    River_width = 100147
    Flow_routing_time_constant = 100148
    _1st_order_rate_AD_model = 100149
    _1st_order_rate_WQ_model = 100150
    Erosion_deposition_coefficient = 100151
    Shear_stress = 100152
    Dispersion_coefficient = 100153
    Dispersion_factor = 100154
    Sediment_volume_per_length_unit = 100155
    Latitude_longitude = 100157
    Specific_gravity = 100158
    Transmission_coefficient = 100159
    Reflection_coefficient = 100160
    Friction_factor = 100161
    Radiation_intensity = 100162
    Duration = 100163
    Respiration_production_per_area = 100164
    Respiration_production_per_volume = 100165
    Sediment_depth = 100166
    Angle_of_repose = 100167
    Half_order_rate_WQ_model = 100168
    Rearation_constant = 100169
    Deposition_rate = 100170
    BOD_at_river_bed = 100171
    Crop_demand = 100172
    Irrigated_area = 100173
    Livestock_demand = 100174
    Number_of_livestock = 100175
    Total_Gas = 100176
    Ground_water_abstraction = 100177
    Melting_coefficient = 100178
    Rain_melting_coefficient_per_degree_per_time = 100179
    Elevation = 100180
    Cross_section_X_data = 100181
    Vegetation_height = 100182
    Geographical_coordinate = 100183
    Angle = 100184
    ItemGeometry0D = 100185
    ItemGeometry1D = 100186
    ItemGeometry2D = 100187
    ItemGeometry3D = 100188
    Temperature_lapse_rate = 100189
    Correction_of_precipitation = 100190
    Temperature_correction = 100191
    Precipitation_correction = 100192
    Max_Water = 100193
    Lower_Baseflow = 100194
    Mass_flux = 100195
    Pressure2 = 100196
    Turbulent_kinetic_energy = 100197
    Dissipation_of_TKE = 100198
    Salt_Flux = 100199
    Temperature_Flux = 100200
    Concentration_Non_Dim = 100201
    Latent_Heat = 100202
    Heat_Flux = 100203
    Specific_Heat = 100204
    Visibility = 100205
    Ice_thickness = 100206
    Structure_geometry___time = 100207
    Discharge___time = 100208
    Fetch_length = 100209
    Rubble_mound = 100210
    Grid_Spacing = 100211
    TimeStep = 100212
    Length_Scale = 100213
    Erosion_Coefficient_Factor = 100214
    Friction_Coefficient = 100215
    Transition_Rate = 100216
    Distance = 100217
    Time_Correction_At_Noon = 100218
    Critical_Velocity = 100219
    Light_Extinction_Background = 100220
    Particle_Production_Rate = 100221
    First_Order_Grazing_Rate_Dependance = 100222
    Resuspension_Rate = 100223
    Adsorption_Coefficient = 100224
    Desorption_Coefficient = 100225
    Sedimentation_Velocity = 100226
    Boundary_Layer_Thickness = 100227
    Diffusion_Coefficient = 100228
    Bioconcentration_Factor = 100229
    Fcoli_Concentration = 100230
    Specific_Discharge = 100231
    Precipitation = 100232
    Specific_Precipitation = 100233
    Power = 100234
    Conveyance_Loss = 100235
    Infiltration_Flux = 100236
    Evaporation_Flux = 100237
    Ground_Water_Abstraction_Flux = 100238
    Fraction = 100239
    Yield_Factor = 100240
    Specific_Solute_Flux_per_Area = 100241
    Current_Speed = 100242
    Current_Direction = 100243
    Current_Magnitude = 100244
    First_Order_Piston_Position = 100245
    Subharmonic_Piston_Position = 100246
    Superharmonic_Piston_Position = 100247
    First_Order_Flap_Position = 100248
    Subharmonic_Flap_Position = 100249
    Superharmonic_Flap_Position = 100250
    Length_Zero_Crossing = 100251
    Time_Zero_Crossing = 100252
    Length_Logged_Data = 100253
    Force_Logged_Data = 100254
    Speed_Logged_Data = 100255
    Volume_Flow_Logged_Data = 100256
    _2D_Surface_Elevation_Spectrum = 100257
    _3D_Surface_Elevation_Spectrum = 100258
    Directional_Spreading_Function = 100259
    Auto_Spectrum = 100260
    Cross_Spectrum = 100261
    Coherence_Spectrum = 100262
    Coherent_Spectrum = 100263
    Frequency_Response_Spectrum = 100264
    Phase_Spectrum = 100265
    FIR_Filter_coefficient = 100266
    Fourier_a_Coefficient = 100267
    Fourier_b_Coefficient = 100268
    u_velocity_component = 100269
    v_velocity_component = 100270
    w_velocity_component = 100271
    Bed_Thickness = 100272
    Dispersion_Velocity_Factor = 100273
    Wind_speed = 100274
    Shore_Current_Zone = 100275
    Depth_of_Wind = 100276
    Emulsification_Constant_K1 = 100277
    Emulsification_Constant_K2 = 100278
    Light_Extinction = 100279
    Water_Depth = 100280
    Reference_settling_velocity = 100281
    Phase_Error = 100282
    Level_Amplitude_Error = 100283
    Discharge_Amplitude_Error = 100284
    Level_Correction = 100285
    Discharge_Correction = 100286
    Level_Simulated = 100287
    Discharge_Simulated = 100288
    Summ_Q_Corrected = 100289
    Time_Scale = 100290
    Sponge_Coefficient = 100291
    Porosity_Coefficient = 100292
    Filter_Coefficient = 100293
    Skewness = 100294
    Asymmetry = 100295
    Atiltness = 100296
    Kurtosis = 100297
    Auxiliary_variable_w = 100298
    Roller_Thickness = 100299
    Line_Thickness = 100300
    Marker_Size = 100301
    Roller_Celerity = 100302
    Encroachment_offset = 100303
    Encroachment_position = 100304
    Encroachment_width = 100305
    Conveyance_reduction = 100306
    Water_level_change = 100307
    Energy_level_change = 100308
    Horizontal_particle_velocity = 100309
    Vertical_particle_velocity = 100310
    Area_fraction = 100311
    Catchment_slope = 100312
    Average_length = 100313
    PE = 100314
    Inverse_exponent = 100315
    Time_shift = 100316
    Attenuation = 100317
    Population = 100318
    Industrial_output = 100319
    Agricultural_area = 100320
    Population_usage = 100321
    Industrial_use = 100322
    Agricultural_usage = 100323
    Layer_Thickness = 100324
    Snow_Depth = 100325
    Snow_Cover_Percentage = 100326
    Pressure_Head = 100353
    Crop_Coefficient = 100354
    Aroot_Kristensen_and_Jensen = 100355
    C1_Kristensen_and_Jensen = 100356
    C2_Kristensen_and_Jensen = 100357
    C3_Kristensen_and_Jensen = 100358
    Irrigation_Demand = 100359
    Transmissivity = 100360
    Darcy_Velocity = 100361
    Leakage_Coeff_Drain_Time_Const = 100362
    Conductance = 100363
    Height_Above_Ground = 100364
    Pumping_Rate = 100365
    Depth_Below_Ground = 100366
    Cell_Height = 100367
    Head_Gradient = 100368
    Ground_Water_Flow_Velocity = 100369
    Grid_Codes = 100370
    Drainage_Time_Constant = 100371
    Head_Elevation = 100372
    Length_Error = 100373
    Storage_Coefficient = 100374
    Specific_Yield = 100375
    Exchange_Rate = 100376
    Volumetric_Water_Content = 100377
    Storage_Change_Rate = 100378
    Seepage = 100379
    Root_Depth = 100380
    Rill_Depth = 100381
    Logical = 100382
    Leaf_Area_Index = 100383
    Irrigation_Rate = 100384
    Irrigation_Index = 100385
    Interception = 100386
    Evapotranspiration_Rate = 100387
    Erosion_Surface_Load = 100388
    Erosion_Concentration = 100389
    Epsilon_UZ = 100390
    Drainage = 100391
    Deficit = 100392
    Crop_Yield = 100393
    Crop_Type = 100394
    Crop_Stress = 100395
    Crop_Stage = 100396
    Crop_Loss = 100397
    Crop_Index = 100398
    Age = 100399
    Conductivity = 100400
    Print_Scale_Equivalence = 100401
    Concentration_1 = 100402
    Concentration_2 = 100403
    Concentration_3 = 100404
    Concentration_4 = 100405
    Sediment_diameter = 100406
    Mean_Wave_Direction = 100407
    Flow_Direction_1 = 100408
    Air_Pressure = 100409
    Decay_Factor = 100410
    Sediment_Bed_Density = 100411
    Dispersion_Coefficient = 100412
    Velocity_Profile = 100413
    Habitat_Index = 100414
    Angles = 100415
    Hydraulic_Length = 100416
    SCS_Catchment_slope = 100417
    Turbidity_FTU = 100418
    Turbidity_MgPerL = 100419
    Bacteria_Flow = 100420
    Bed_Distribution = 100421
    Surface_Elevation_at_Paddle = 100422
    Unit_Hydrograph_Response = 100423
    Transfer_Rate = 100424
    Return_period = 100425
    Constant_Settling_Velocity = 100426
    Deposition_Concentration_Flux = 100427
    Settling_Velocity_Coefficient = 100428
    Erosion_Coefficient = 100429
    Volume_Flux = 100430
    Precipitation_Rate = 100431
    Evaporation_Rate = 100432
    Co_Spectrum = 100433
    Quad_Spectrum = 100434
    Propagation_Direction = 100435
    Directional_Spreading = 100436
    Mass_per_Unit_Area = 100437
    Incident_Spectrum = 100438
    Reflected_Spectrum = 100439
    Reflection_Function = 100440
    Bacteria_Flux = 100441
    Head_Difference = 100442
    Energy = 100443
    Directional_Standard_Deviation = 100444
    Rainfall_Depth = 100445
    Ground_Water_Abstraction_Depth = 100446
    Evapo_Transpiration = 100447
    Longitudinal_Infiltration = 100448
    Pollutant_Load = 100449
    Pressure = 100450
    Cost_per_Time = 100451
    Mass = 100452
    Mass_per_Time = 100453
    Mass_per_Area_per_Time = 100454
    Kd = 100455
    Porosity = 100456
    Half_Life = 100457
    Dispersivity = 100458
    Friction_Coefficient_cfw = 100459
    Wave_amplitude = 100460
    Sediment_grain_diameter = 100461
    Sediment_spill = 100463
    Number_of_Particles = 100464
    Ellipsoidal_height = 100500
    Cloudiness = 100501
    Probability = 100502
    Activity_of_Dispersant = 100503
    Dredge_Rate = 100504
    Dredge_Spill = 100505
    Clearness_Coefficient = 100506
    Coastline_Orientation = 100507
    Reduction_Factor = 100508
    Active_Beach_Height = 100509
    Update_Period = 100510
    Accumulated_Erosion = 100511
    Erosion_Rate = 100512
    Non_dimensional_Transport = 100513
    Local_coordinate = 100514
    Radii_of_Gyration = 100515
    Percentage = 100516
    Line_Capacity = 100517
    Undefined = 999
    Diverted_discharge = 110001
    Demand_carry_over_fraction = 110002
    Groundwater_demand = 110003
    Dam_crest_level = 110004
    Seepage_flux = 110005
    Seepage_fraction = 110006
    Evaporation_fraction = 110007
    Residence_time = 110008
    Owned_fraction_of_inflow = 110009
    Owned_fraction_of_volume = 110010
    Reduction_level = 110011
    Reduction_threshold = 110012
    Reduction_fraction = 110013
    Total_losses = 110014
    Counts_per_liter = 110015
    Assimilative_Capacity = 110016
    Still_Water_Depth = 110017
    Total_Water_Depth = 110018
    Maximum_wave_height = 110019
    Ice_concentration = 110020
    Wind_friction_speed = 110021
    Roughness_length = 110022
    Drag_coefficient = 110023
    Charnock_constant = 110024
    Breaking_parameter_Gamma = 110025
    Threshold_period = 110026
    Courant_number = 110027
    Time_step_factor = 110028
    Element_length = 110029
    Element_area = 110030
    Roller_angle = 110031
    Rate_of_bed_level_change = 110032
    Bed_level_change = 110033
    Sediment_transport_direction = 110034
    Wave_action_density = 110035
    Zero_moment_of_wave_action = 110036
    First_moment_of_wave_action = 110037
    Bed_Mass = 110038
    Water_Quality = 110039
    Status = 110040
    Setting = 110041
    Reaction_Rate = 110042
    Fast_Runoff_Discharge = 110043
    Slow_Runoff_Discharge = 110044
    Average_Sediment_Transport_per_length_unit = 110045
    Valve_Setting = 110046
    Wave_energy_density = 110047
    Wave_energy_distribution = 110048
    Wave_energy = 110049
    Radiation_Melting_Coefficient = 110050
    Rain_melting_coefficient = 110051
    Friction = 110052
    Wave_action_density_rate = 110053
    Element_area_long_lat = 110054
    Electric_Current = 110100
    Heat_Flux_Resistance = 110200
    Absolute_Humidity = 110210
    Length = 110220
    Area = 110225
    Volume = 110230
    Element_Volume = 110231
    Wave_Power = 110232
    Moment_of_Inertia = 110233
    Topography = 110234
    Scour_Depth = 110235
    Scour_Width = 110236
    Cost_per_Volume = 110237
    Cost_per_Energy = 110238
    Cost_per_Mass = 110239
    Application_Intensity = 110240
    Cost = 110241
    Voltage = 110242
    Normal_Velocity = 110243
    Gravity = 110244
    Vessel_Displacement = 110245
    Hydrostatic_Matrix = 110246
    Wave_Number = 110247
    Radiation_Potential = 110248
    Added_Mass_TT = 110249
    Radiation_Damping = 110250
    Frequency = 110251
    Sound_Exposure_Level = 110252
    Transmission_Loss = 110253
    pH = 110254
    Acoustic_Attenuation = 110255
    Sound_Speed = 110256
    Leakage = 110257
    Height_Above_Keel = 110258
    Submerged_Mass = 110259
    Deflection = 110260
    Linear_Damping_Coefficient = 110261
    Quadratic_Damping_Coefficient = 110262
    Damping_TT = 110263
    RAO_Motion = 110264
    RAO_Rotation = 110265
    Added_Mass_Coefficient = 110266
    Electrical_Conductivity = 110267
    Added_Mass_TR = 110268
    Added_Mass_RT = 110269
    Added_Mass_RR = 110270
    Damping_TR = 110271
    Damping_RT = 110272
    Damping_RR = 110273
    Fender_Force = 110274
    Force = 110275
    Moment = 110276
    Reduced_Pollutant_Load = 110277
    Size_and_Position = 110278
    Frame_Rate = 110279
    Dynamic_Viscosity = 110280
    Grid_Rotation = 110281
    Agent_Density = 110282
    Emitter_Coefficient = 110283
    Pipe_Diameter = 110284
    Speed = 110285
    Velocity = 110286
    Direction = 110287
    Displacement = 110288
    Position = 110289
    Rotation = 110290
    Torque = 110291
    Overtopping = 110292
    Flow_Rate = 110293
    Acceleration = 110294
    Dimensionless_Acceleration = 110295
    Time = 110296
    Resistance = 110297
    Amount_of_Substance = 110298
    Molar_Concentration = 110299
    Molal_Concentration = 110300
    Suspended_sediment_load_per_area = 110301
    Bollard_Force = 110302
    Discharge_per_Pressure = 110303
    RotationalSpeed = 110304
    Infiltration_per_Area = 110305
    Mass_per_Length_per_Time = 110306
    NearBedLoad_per_Length = 110307
    Substance_per_UnitArea = 110308
    AccNearBedLoad_per_Length = 110309
    ThermalConductivity = 110310
    DirectionalVariance = 110311
    SpecificDissipationRate = 110312

    def __init__(self, code: int) -> None:
        self.code = code

    @property
    def display_name(self) -> str:
        """Display friendly name."""
        name = self.name
        name = name.replace("_", " ")
        return name

    def __repr__(self) -> str:
        return self.display_name

    @property
    def units(self) -> list[EUMUnit]:
        """List valid units for this EUM type."""
        temp = _unit_list(self.code).values()
        return [EUMUnit(value) for value in temp]

    @staticmethod
    def search(pattern: str) -> list[EUMType]:
        temp = _type_list(pattern).keys()
        return [EUMType(key) for key in temp]


class EUMUnit(IntEnum):
    """EUM unit.

    Examples
    --------
    ```{python}
    import mikeio
    mikeio.EUMUnit.degree_Kelvin
    ```

    """

    meter = 1000
    kilometer = 1001
    centimeter = 1007
    millimeter = 1002
    feet = 1003
    feet_US = 1014
    inch = 1004
    inch_US = 1013
    mile = 1005
    mile_US = 1016
    yard = 1006
    yard_US = 1015
    meter_pow_3_per_sec = 1800
    meter_pow_3_per_min = 1816
    feet_pow_3_per_sec = 1801
    feet_pow_3_per_min = 1817
    feet_pow_3_per_day = 1814
    feet_pow_3_per_year = 1815
    meter_pow_3_per_day = 1810
    meter_pow_3_per_year = 1805
    acre_feet_per_day = 1804
    meter_pow_3_per_hour = 1821
    gallon_per_min = 1818
    liter_per_minute = 1820
    liter_per_sec = 1819
    Mgal_per_day = 1803
    MgalUK_per_day = 1823
    Ml_per_day = 1802
    meter_per_sec = 2000
    feet_per_sec = 2002
    miles_per_hour = 2020
    km_per_hour = 2021
    knot = 2019
    degree = 2401
    degree_Celsius = 2800
    delta_degree_Celsius = 2900
    degree_Fahrenheit = 2801
    delta_degree_Fahrenheit = 2901
    degree_Kelvin = 2802
    mu_g_per_meter_pow_3 = 2201
    mg_per_meter_pow_3 = 2202
    gram_per_meter_pow_3 = 2203
    kg_per_meter_pow_3 = 2200
    mu_g_per_liter = 2204
    mg_per_liter = 2205
    gram_per_liter = 2206
    pound_per_feet_pow_3 = 2207
    ton_per_meter_pow_3 = 2208
    million_per__100_ml = 3000
    per__100_ml = 3001
    per_liter = 3002
    __ = 99000
    feet_US_pow_3_per_sec = 1831
    feet_US_pow_3_per_day = 1833
    feet_US_pow_3_per_year = 1834
    feet_US_pow_3_per_min = 1832
    yard_US_pow_3_per_sec = 1835
    yard_pow_3_per_sec = 1830
    meter_per_day = 2006
    feet_US_per_sec = 2036
    percent = 99001
    percent_per_day = 2601
    cm_per_sec = 2023
    mm_per_sec = 2017
    inch_per_sec = 2011
    inch_US_per_sec = 2035
    hours_per_day = 99003
    watt_per_meter_pow_2 = 6600
    joule_per_meter_pow_2_per_day = 5703
    kJ_per_meter_pow_2_per_hour = 5700
    kJ_per_meter_pow_2_per_day = 5701
    MJ_per_meter_pow_2_per_day = 5702
    PSU = 6200
    per_thousand = 99002
    meter_pow_2 = 3200
    feet_pow_2 = 3203
    meter_pow__1_per_3__per_sec = 3600
    feet_pow__1_per_3__per_sec = 3601
    sec_per_meter_pow__1_per_3_ = 3800
    sec_per_feet_pow__1_per_3_ = 3801
    meter_pow__1_per_2__per_sec = 4000
    feet_pow__1_per_2__per_sec = 4001
    feet_US_pow__1_per_2__per_sec = 4002
    meter_pow_3 = 1600
    liter = 1601
    megaliter = 1609
    km_pow_3 = 1606
    acre_feet = 1607
    feet_pow_3 = 1603
    gallon = 1604
    megagallon = 1608
    _10_pow_6meter_pow_3 = 1610
    acre = 3202
    kilogram = 1200
    gram = 1201
    milligram = 1202
    microgram = 1203
    ton = 1204
    pound = 1207
    kiloton = 1205
    megaton = 1206
    ounce = 1209
    per_kg = 1250
    per_gram = 1251
    per_mg = 1252
    per_mu_g = 1253
    per_ton = 1254
    per_kiloton = 1255
    per_megaton = 1256
    per_pound = 1257
    per_ton_US = 1258
    per_ounce = 1259
    kg_per_sec = 4200
    gram_per_sec = 4203
    mg_per_sec = 4202
    mu_g_per_sec = 4201
    pound_per_sec = 4215
    yard_pow_3 = 1615
    yard_US_pow_3 = 1614
    liter_per_day = 1836
    mm_per_hour = 2001
    mm_per_day = 2004
    inch_per_hour = 2016
    cm_per_hour = 2018
    feet_per_day = 2009
    inch_per_min = 2014
    mu_m_per_sec = 2031
    millimeter_per_day = 4801
    inch_per_day = 4802
    liter_per_sec_per_ha = 2030
    mm_per_year = 2040
    meter_pow_3_per_meter = 3201
    feet_pow_3_per_feet = 3207
    feet_US_pow_3_per_feet_US = 3214
    yard_pow_3_per_yard = 3213
    yard_US_pow_3_per_yard_US = 3212
    kg_per_hour = 4204
    gram_per_meter_pow_2 = 4400
    gram_per_meter_pow_2_per_day = 4500
    gram_per_meter_pow_3_per_hour = 4600
    ton_US = 1208
    kg_per_meter = 4401
    meter_pow_3_per_sec_per_meter = 4700
    meter_pow_2_per_sec = 4702
    feet_pow_2_per_sec = 4704
    cm_pow_3_per_sec_per_cm = 4721
    mm_pow_3_per_sec_per_mm = 4722
    feet_pow_3_per_sec_per_feet = 4718
    feet_US_pow_3_per_sec_per_feet_US = 4723
    inch_pow_3_per_sec_per_inch = 4724
    inch_US_pow_3_per_sec_per_inch_US = 4725
    yard_US_pow_3_per_sec_per_yard_US = 4726
    yard_pow_3_per_sec_per_yard = 4727
    meter_pow_3_per_year_per_meter = 4701
    second = 1400
    hertz = 2602
    day = 1403
    liter_per_sec_per_km_pow_2 = 2003
    acre_feet_per_sec_per_acre = 2005
    feet_pow_3_per_sec_per_mile_pow_2 = 2007
    watt = 4900
    kilowatt = 4901
    megawatt = 4902
    gigawatt = 4903
    radian = 2400
    per_meter = 5000
    per_feet = 5003
    per_feet_US = 5005
    meter_pow_3_per_sec_pow_2 = 6800
    meter_pow_2_sec_per_rad = 5200
    feet_pow_2_sec_per_rad = 5213
    meter_pow_2_per_rad = 5201
    feet_pow_2_per_rad = 5215
    meter_pow_2_sec = 5202
    feet_pow_2_sec = 5217
    square_meter_per_sec = 9400
    square_feet_per_sec = 9401
    Integer = 99013
    hectare = 3204
    km_pow_2 = 3205
    mile_pow_2 = 3206
    millifeet = 1010
    kg_per_day = 4205
    kg_per_year = 4207
    hour = 1402
    per_hour = 2603
    per_day = 2600
    per_sec = 2605
    gram_per_meter_pow_2_per_sec = 4501
    kg_per_meter_pow_2_per_sec = 4503
    newton_per_meter_pow_2 = 5400
    kN_per_meter_pow_2 = 5401
    pound_per_feet_per_sec_pow_2 = 5402
    newton_per_meter_pow_3 = 5500
    kN_per_meter_pow_3 = 5501
    kilogram_M2 = 5550
    poundSqrFeet = 5551
    gram_per_meter_pow_3_per_day = 4601
    _mg_per_l__pow__1_per_2__per_day = 5300
    _mg_per_l__pow__1_per_2__per_hour = 5301
    meter_per_hour = 2008
    gal_per_day_per_head = 1806
    liter_per_day_per_head = 1807
    meter_pow_3_per_sec_per_head = 1808
    mm_per_month = 2010
    mm_per_C_per_day = 5800
    mm_per_C_per_hour = 5801
    inch_per_F_per_day = 5802
    inch_per_F_per_hour = 5803
    _per_C_per_day = 5900
    _per_C_per_hour = 5901
    _per_F_per_day = 5902
    _per_F_per_hour = 5903
    Celsius_per__100meter = 6000
    Celsius_per__100feet = 6001
    Fahrenheit_per__100feet = 6003
    Fahrenheit_per__100meter = 6002
    percent_per__100meter = 5001
    percent_per__100feet = 5002
    gram_per_day = 4206
    ton_per_sec = 4220
    pascal = 6100
    hectopascal = 6101
    millibar = 6108
    meter_pow_2_per_sec_pow_2 = 6400
    feet_pow_2_per_sec_pow_2 = 4720
    meter_pow_2_per_sec_pow_3 = 6401
    feet_pow_2_per_sec_pow_3 = 6402
    PSU_meter_pow_3_per_sec = 6300
    non_dim_meter_pow_3_per_sec = 6302
    PSU_feet_pow_3_per_sec = 6303
    C_meter_pow_3_per_sec = 6301
    F_feet_pow_3_per_sec = 6304
    mg_per_gram = 99007
    ml_per_liter = 99018
    per_million = 99020
    mg_per_kg = 99008
    mu_l_per_liter = 99019
    joule_per_kg = 6500
    joule_kg_per_K = 6700
    meter_per_min = 2012
    feet_per_min = 2013
    feet_per_hour = 2015
    feet_pow_3_per_sec_pow_2 = 6801
    nautical_mile = 1009
    minute = 1401
    month = 1405
    year = 1404
    gram_per_meter_pow_3_per_sec = 4602
    meter_pow_3_per_gram = 6900
    liter_per_gram = 6901
    meter_pow_3_per_ha_per_sec = 2027
    meter_pow_3_per_ha_per_hour = 2026
    meter_pow_3_per_ha_per_day = 2025
    feet_pow_3_per_acre_per_sec = 2024
    feet_pow_3_per_acre_per_hour = 2041
    feet_pow_3_per_acre_per_day = 2042
    liter_per_min_per_ha = 2029
    gallon_per_min_per_acre = 2028
    horsepower = 4904
    kg_per_ha_per_hour = 4502
    kg_per_ha_per_day = 4504
    pound_per_feet_pow_2_per_sec = 4507
    pound_per_acre_per_day = 4505
    newton = 7000
    meter_pow_2_per_hertz = 7100
    feet_pow_2_per_hertz = 7103
    meter_pow_2_per_hertz_per_deg = 7101
    meter_pow_2_per_hertz_per_rad = 7102
    feet_pow_2_per_hertz_per_deg = 7104
    feet_pow_2_per_hertz_per_rad = 7105
    kg_per_sec_pow_2 = 8100
    meter_pow_2_per_kg = 9100
    person = 99004
    currency_per_year = 2604
    liter_per_person_per_day = 1809
    meter_pow_3_per_currency = 1611
    meter_pow_3_per_km_pow_2_per_day = 4803
    per_inch = 5004
    meter_pow_2_per_hour = 4708
    meter_pow_2_per_day = 4709
    feet_pow_2_per_hour = 4710
    feet_pow_2_per_day = 4711
    feet_US_per_day = 2037
    inch_US_per_hour = 2038
    inch_US_per_min = 2039
    gallon_per_sec = 1811
    gallon_per_day = 1812
    gallon_per_year = 1813
    kg_per_meter_pow_2 = 4402
    meter_pow_3_per_meter_pow_3 = 99011
    liter_per_meter_pow_3 = 99012
    kg_per_ha = 4403
    kg_per_km_pow_2 = 4406
    gram_per_km_pow_2 = 4408
    gram_per_ha = 4410
    ton_per_meter_pow_2 = 2210
    ton_per_km_pow_2 = 4407
    ton_per_ha = 4409
    pound_per_feet_pow_2 = 2209
    pound_per_acre = 4405
    pound_per_mile_pow_2 = 4411
    kg_per_acre = 4412
    kg_per_feet_pow_2 = 4413
    kg_per_mile_pow_2 = 4414
    ton_per_acre = 4415
    ton_per_feet_pow_2 = 4416
    ton_per_mile_pow_2 = 4417
    gram_per_acre = 4418
    gram_per_feet_pow_2 = 4419
    gram_per_mile_pow_2 = 4420
    pound_per_ha = 4421
    pound_per_meter_pow_2 = 4422
    pound_per_km_pow_2 = 4423
    pound_per_feet_US_pow_3 = 2214
    pound_per_yard_US_pow_3 = 2212
    pound_per_yard_pow_3 = 2213
    gram_per_gram = 99005
    gram_per_kg = 99006
    mu_g_per_gram = 99009
    kg_per_kg = 99010
    micrometer = 1008
    _10_pow_9_per_day = 2606
    _10_pow_12_per_year = 2607
    meter_pow_3_per_sec_per__10mm = 4706
    feet_pow_3_per_sec_per_inch = 4707
    mg_per_meter_pow_2 = 4404
    mu_g_per_meter_pow_2 = 2211
    _1000_per_meter_pow_2_per_day = 3401
    _per_meter_pow_2_per_sec = 3402
    terajoule = 5604
    gigajoule = 5603
    megajoule = 5602
    kilojoule = 5601
    joule = 5600
    petajoule = 5607
    exajoule = 5608
    kilowatt_hour = 5605
    megawatt_hour = 5609
    gigawatt_hour = 5610
    liter_per_meter_pow_2 = 1011
    meter_pow_3_per_hour_per_meter = 7503
    meter_pow_3_per_day_per_meter = 7504
    feet_pow_3_per_hour_per_feet = 4719
    m3_per_hour_per_M = 4730
    m3_per_day_per_M = 4731
    feet_pow_3_per_day_per_feet = 4732
    galUK_per_day_per_feet = 4712
    gallon_per_day_per_feet = 4713
    gallon_per_min_per_feet = 4714
    liter_per_day_per_meter = 4715
    liter_per_minute_per_meter = 4716
    liter_per_second_per_meter = 4717
    liter_per_min_per_meter = 7501
    liter_per_sec_per_meter = 7500
    gram_per_min = 4208
    pound_per_day = 4212
    pound_per_hour = 4213
    pound_per_min = 4214
    pound_per_year = 4217
    ton_per_year = 4218
    MetresWater = 6105
    FeetWater = 6106
    kilopascal = 6102
    megapascal = 6104
    psi = 6103
    bar = 6107
    decibar = 6110
    kg_per_meter_pow_2_per_day = 4506
    meter_pow_3_per_mg = 6902
    meter_pow_3_per_mu_g = 6903
    ton_per_day = 4219
    millimeterD50 = 1012
    millisecond = 1406
    feet_US_pow_2 = 3208
    yard_US_pow_2 = 3209
    mile_US_pow_2 = 3210
    acre_US = 3211
    liter_per_meter = 3215
    milliliter = 1602
    milligallon = 1605
    gallonUK = 1612
    megagallonUK = 1613
    galUK_per_day = 1822
    feet_pow_3_per_PE_per_day = 1824
    meter_pow_3_per_PE_per_day = 1825
    galUK_per_sec = 1826
    galUK_per_year = 1827
    galUK_per_PE_per_day = 1828
    ydUS3_per_sec = 1829
    acre_feet_per_day_per_acre = 2022
    Mgal_per_day_per_acre = 2032
    MgalUK_per_day_per_acre = 2033
    Ml_per_day_per_ha = 2034
    liter_per_hour_per_ha = 2043
    liter_per_day_per_ha = 2044
    meter_per_sec_pow_2 = 2100
    feet_per_sec_pow_2 = 2101
    pound_per_feet_US_pow_2 = 2215
    ounce_per_cubic_feet = 2216
    ounce_per_cubic_feet_US = 2217
    ounce_per_Yard3 = 2218
    ounce_per_yard_US3 = 2219
    ounce_per_square_feet = 2220
    ounce_per_square_feet_US = 2221
    kg_per_meter_per_sec = 2300
    Pascal_second = 2301
    kilogram_per_meter_per_day = 2302
    gram_per_Meter_per_Day = 2303
    gram_per_Km_per_Day = 2304
    pound_per_Feet_per_Day = 2305
    pound_per_FeetUS_per_Day = 2306
    ounce_per_Feet_per_Day = 2307
    ounce_per_FeetUS_per_Day = 2308
    kilogram_per_Yard_per_Second = 2309
    kilogram_per_Feet_per_Second = 2310
    pound_per_Yard_per_Second = 2311
    pound_per_Feet_per_Second = 2312
    degree50 = 2402
    degree_pow_2 = 2403
    radian_pow_2 = 2404
    degree_per_meter = 2500
    radian_per_meter = 2501
    degree_per_sec = 2510
    radian_per_sec = 2511
    meter_pow_2_per_sec_per_ha = 2608
    feet_pow_2_per_sec_per_acre = 2609
    rev_per_min = 2610
    percent_per_hour = 2611
    percent_per_sec = 2613
    revolution_per_second = 2614
    revolution_per_hour = 2615
    _per_degree_C = 2850
    _per_degree_F = 2851
    per_meter_pow_3 = 3003
    per_ml = 3004
    per_feet_pow_3 = 3005
    per_gal = 3006
    per_mgal = 3007
    per_km_pow_3 = 3008
    per_acre_feet = 3009
    per_Mgal = 3010
    per_Ml = 3011
    per_gallonUK = 3012
    per_MgalUK = 3013
    per_yard_US_pow_3 = 3014
    per_yard_pow_3 = 3015
    sec_per_meter = 3100
    Einstein_per_meter_pow_2_per_day = 3400
    kg_per_PE_per_day = 4209
    kg_per_min = 4210
    pound_per_PE_per_day = 4216
    mg_per_ha = 4424
    mg_per_km_pow_2 = 4425
    mg_per_acre = 4426
    mg_per_feet_pow_2 = 4427
    mg_per_mile_pow_2 = 4428
    pound_per_meter = 4429
    ton_per_meter = 4430
    pound_per_Feet = 4431
    pound_per_Yard = 4432
    pound_per_FeetUS = 4433
    pound_per_YardUS = 4434
    ounce_per_Feet = 4435
    ounce_per_Yard = 4436
    ounce_per_FeetUS = 4437
    ounce_per_YardUS = 4438
    kilogram_per_Yard = 4439
    kilogram_per_Feet = 4440
    mg_per_liter_per_day = 4603
    yard_pow_3_per_year_per_yard = 4728
    yard_US_pow_3_per_year_per_yard_US = 4729
    per_inch_US = 5006
    meter_pow_3_per_sec_2 = 5100
    meter_pow_2_per_deg = 5203
    meter_pow_2_sec_pow_2_per_rad = 5204
    meter_pow_2_per_sec_per_rad = 5205
    meter_pow_2_sec_per_deg = 5206
    meter_pow_2_sec_pow_2_per_deg = 5207
    meter_pow_2_per_sec_per_deg = 5208
    feet_pow_2_per_sec_per_rad = 5209
    feet_pow_2_per_sec_per_deg = 5210
    feet_pow_2_sec_pow_2_per_rad = 5211
    feet_pow_2_sec_pow_2_per_deg = 5212
    feet_pow_2_sec_per_deg = 5214
    feet_pow_2_per_deg = 5216
    kg_meter_pow_2 = 7060
    watt_second = 5606
    per_joule = 5650
    per_kJ = 5651
    per_MJ = 5652
    per_GJ = 5653
    per_TJ = 5654
    per_PJ = 5655
    per_EJ = 5656
    per_kWh = 5657
    per_Ws = 5658
    per_MWh = 5659
    per_GWh = 5660
    mm_per__kJ_per_meter_pow_2_ = 5710
    mm_per__MJ_per_meter_pow_2_ = 5711
    micropascal = 6109
    dB_re_1_mu_Pa_pow_2_sec = 6150
    dB_per_lambda = 6160
    meter_pow_2_per_sec_pow_3_per_rad = 6403
    feet_pow_2_per_sec_pow_3_per_rad = 6404
    AFD_per_sec = 6802
    IMGD_per_sec = 6803
    MGD_per_sec = 6804
    GPM_per_sec = 6805
    meter_pow_3_per_day_per_sec = 6806
    meter_pow_3_per_hour_per_sec = 6807
    Ml_per_day_per_sec = 6808
    liter_per_min_per_sec = 6809
    liter_per_sec_pow_2 = 6810
    kilonewton = 7001
    meganewton = 7002
    millinewton = 7003
    kg_meter = 7050
    kg_meter_per_sec = 7070
    kg_meter_pow_2_per_sec = 7080
    meter_pow_2_per_hertz_pow_2 = 7200
    meter_pow_2_per_hertz_pow_2_per_deg = 7201
    meter_pow_2_per_hertz_pow_2_per_rad = 7202
    Ml_per_day_per_meter = 7502
    feet_pow_3_per_sec_per_psi = 7505
    gallon_per_min_per_psi = 7506
    Mgal_per_day_per_psi = 7507
    MgalUK_per_day_per_psi = 7508
    acre_feet_per_day_per_psi = 7509
    m3_per_hour_per_bar = 7510
    _per_meter_per_sec = 9200
    meter_per_sec_per_ha = 9201
    feet_per_sec_per_acre = 9202
    per_meter_pow_2 = 9300
    per_acre = 9301
    per_hectare = 9302
    per_km_pow_2 = 9303
    per_cubic_meter = 9350
    currency_per_meter_pow_3 = 9351
    currency_per_feet_pow_3 = 9352
    per_watt = 9600
    newton_meter = 9700
    kilonewton_meter = 9701
    meganewton_meter = 9702
    newton_millimeter = 9703
    newton_meter_second = 9800
    newton_per_meter_per_sec = 9900
    mole = 12000
    millimole = 12001
    micromole = 12002
    nanomole = 12003
    mole_per_liter = 12020
    mmol_per_liter = 12021
    mu_mol_per_liter = 12022
    nmol_per_liter = 12023
    mole_per_meter_pow_3 = 12024
    mmol_per_meter_pow_3 = 12025
    mu_mol_per_meter_pow_3 = 12026
    mole_per_kg = 12040
    mmol_per_kg = 12041
    mu_mol_per_kg = 12042
    nmol_per_kg = 12043
    mole_per_m2 = 12060
    millimole_per_m2 = 12061
    micromole_per_m2 = 12062
    nanomole_per_m2 = 12063
    meter_per_meter = 99014
    per_minute = 99015
    percent_per_min = 2612
    per_month = 99016
    per_year = 99017
    g_acceleration = 99021
    ampere = 99100
    milliampere = 99101
    microampere = 99102
    kiloampere = 99103
    megaampere = 99104
    volt = 99150
    millivolt = 99151
    microvolt = 99152
    kilovolt = 99153
    megavolt = 99154
    ohm = 99180
    kOhm = 99181
    Mohm = 99182
    undefined = 0
    watt_per_meter = 99200
    kW_per_meter = 99201
    MW_per_meter = 99202
    GW_per_meter = 99203
    kW_per_feet = 99204
    watt_per_meter_per_degree_celsius = 99220
    watt_per_feet_per_degree_fahrenheit = 99221
    siemens = 99250
    millisiemens = 99251
    microsiemens = 99252
    siemens_per_meter = 99260
    mS_per_cm = 99261
    mu_S_per_cm = 99262
    kg_per__meter_sec_ = 99263
    cPoise = 99264
    lbf_sec_per_feet_pow_2 = 99265
    pound_per__sec_feet_ = 99266

    def __init__(self, code: int) -> None:
        self.code = code

    @property
    def display_name(self) -> str:
        """Display friendly name."""
        name = self.name
        name = name.replace("_", " ")
        return name

    @property
    def short_name(self) -> str:
        unit_short_names = {
            "kilometer": "km",
            "centimeter": "cm",
            "millimeter": "mm",
            "meter": "m",
            "liter": "l",
            "kilogram": "kg",
            "milligram": "mg",
            "gram": "g",
            "second": "s",
            "sec": "s",
            "hertz": "Hz",
            "_pow_2": "^2",
            "_pow_3": "^3",
            "_per_": "/",
            "_": "*",
        }

        name = self.name
        for key, value in unit_short_names.items():
            name = name.replace(key, value)
        return name

    def __repr__(self) -> str:
        return self.display_name


class ItemInfo:
    """Info for dynamicc items (variables).

    Parameters
    ----------
    name: str or EUMType, optional
    type: EUMType or int, optional
        Default EUMType.Undefined
    unit: EUMUnit or int, optional
        Default unit matching EUMType
    data_value_type: str, optional
        One of the following strings: 'Instantaneous', 'Accumulated', 'StepAccumulated', 'MeanStepBackward',
        'MeanStepForward'. Default: 'Instantaneous'

    Examples
    --------
    ```{python}
    import mikeio
    mikeio.ItemInfo("Viken", mikeio.EUMType.Water_Level)
    ```

    ```{python}
    mikeio.ItemInfo(mikeio.EUMType.Wind_speed)
    ```

    """

    def __init__(
        self,
        name: str | EUMType | None = None,
        itemtype: EUMType | EUMUnit | None = None,
        unit: EUMUnit | None = None,
        data_value_type: Literal[
            "Instantaneous", "Accumulated", "StepAccumulated", "MeanStepBackward"
        ] = "Instantaneous",
    ) -> None:
        # Handle arguments in the wrong place
        if isinstance(name, EUMType):
            if isinstance(itemtype, EUMUnit):
                unit = itemtype

            itemtype = name
            name = name.display_name

        if itemtype is not None:
            if isinstance(itemtype, int):
                itemtype = EUMType(itemtype)

            if not isinstance(itemtype, EUMType):
                raise ValueError(
                    "Invalid type. Type should be supplied as EUMType, e.g. ItemInfo('WL',EUMType.Water_Level, EUMUnit.meter)"
                )
            self.type = itemtype

            if name is None:
                name = itemtype.display_name
        else:
            self.type = EUMType.Undefined

        if unit is not None:
            if isinstance(unit, int):
                unit = EUMUnit(unit)

            if not isinstance(unit, EUMUnit):
                raise ValueError(
                    "Invalid unit. Unit should be supplied as EUMUnit, e.g. ItemInfo('WL',EUMType.Water_Level, EUMUnit.meter)"
                )
            self._unit = unit
        else:
            if self.type == EUMType.Undefined:
                self._unit = EUMUnit.undefined
            else:
                self._unit = self.type.units[0]

        self.data_value_type = to_datatype(data_value_type)

        if not isinstance(name, str):
            raise ValueError("Invalid name, name should be a string")
        self.name: str = name

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ItemInfo):
            return NotImplemented

        # an alternative approach to this long expression is to use dataclasses (Python>=3.7)
        return (
            self.name == other.name
            and self.type == other.type
            and self.unit == other.unit
            and self.data_value_type == other.data_value_type
        )

    def __repr__(self) -> str:
        if self.data_value_type == DataValueType.Instantaneous:
            return f"{self.name} <{self.type.display_name}> ({self.unit.display_name})"
        else:
            return f"{self.name} <{self.type.display_name}> ({self.unit.display_name}) - {self.data_value_type}"

    @property
    def unit(self) -> EUMUnit:
        "Item unit."
        return self._unit

    @unit.setter
    def unit(self, value: EUMUnit) -> None:
        "Set unit."
        if value not in self.type.units:
            raise ValueError(
                f"{value} is not a correct unit for {self.type}. Use {self.type.units}"
            )
        self._unit = value

    @staticmethod
    def from_mikecore_dynamic_item_info(dfsItemInfo: DfsDynamicItemInfo) -> "ItemInfo":
        """Create ItemInfo from a mikecore.DfsDynamicItemInfo object."""
        name = dfsItemInfo.Name
        item = dfsItemInfo.Quantity.Item
        unit = dfsItemInfo.Quantity.Unit
        itemtype = EUMType(item)
        unit = EUMUnit(unit)
        data_value_type = dfsItemInfo.ValueType

        return ItemInfo(name, itemtype, unit, data_value_type)


class ItemInfoList(list):
    def __init__(self, items: Sequence[ItemInfo]):
        super().__init__(items)

    def to_dataframe(self) -> pd.DataFrame:
        data = [
            {"name": item.name, "type": item.type.name, "unit": item.unit.name}
            for item in self
        ]
        return pd.DataFrame(data)


def to_datatype(datatype: str | int | DataValueType) -> DataValueType:
    string_datatype_mapping = {
        "Instantaneous": DataValueType.Instantaneous,
        "Accumulated": DataValueType.Accumulated,
        "StepAccumulated": DataValueType.StepAccumulated,
        "MeanStepBackward": DataValueType.MeanStepBackward,
        "MeanStepForward": DataValueType.MeanStepForward,
        0: DataValueType.Instantaneous,
        1: DataValueType.Accumulated,
        2: DataValueType.StepAccumulated,
        3: DataValueType.MeanStepBackward,
        4: DataValueType.MeanStepForward,
    }

    if isinstance(datatype, str):
        if datatype not in string_datatype_mapping.keys():
            raise InvalidDataValueType

        return string_datatype_mapping[datatype]

    if isinstance(datatype, int):
        if datatype not in string_datatype_mapping.keys():
            raise InvalidDataValueType

        return string_datatype_mapping[datatype]

    if not isinstance(datatype, DataValueType):
        raise ValueError("Data value type not supported")

    return datatype



from __future__ import annotations
from collections.abc import KeysView, ValuesView
from datetime import datetime
from types import SimpleNamespace
from typing import (
    Any,
    ItemsView,
    Mapping,
    MutableMapping,
    Sequence,
)

import pandas as pd


def _merge_dict(a: dict[str, Any], b: Mapping[str, Any]) -> dict[str, Any]:
    """merges dict b into dict a; handling non-unique keys."""
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                _merge_dict(a[key], b[key])
            else:
                ab = list(a[key]) + list(b[key])
                a[key] = PfsNonUniqueList(ab)
        else:
            a[key] = b[key]
    return a


class PfsNonUniqueList(list):
    # TODO do we really need this, regular lists are not unique
    pass


class PfsSection(SimpleNamespace, MutableMapping[str, Any]):
    """Class for reading/writing sections in a pfs file."""

    @staticmethod
    def from_dataframe(df: pd.DataFrame, prefix: str) -> "PfsSection":
        """Create a PfsSection from a DataFrame.

        Parameters
        ----------
        df: dataframe
            data
        prefix: str
            section header prefix

        Examples
        --------
        ```{python}
        import pandas as pd
        import mikeio
        df = pd.DataFrame(dict(station=["Foo", "Bar"],include=[0,1]), index=[1,2])
        df
        ```

        ```{python}
        mikeio.PfsSection.from_dataframe(df,"STATION_")
        ```

        """
        d = {f"{prefix}{idx}": row.to_dict() for idx, row in df.iterrows()}

        return PfsSection(d)

    def __init__(self, dictionary: Mapping[str, Any], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        for key, value in dictionary.items():
            self.__set_key_value(key, value, copy=True)

    def __repr__(self) -> str:
        return "\n".join(self._to_txt_lines())

    def __len__(self) -> int:
        return len(self.__dict__)

    def __iter__(self) -> Any:
        iter(self)

    def __contains__(self, key: Any) -> bool:
        return key in self.keys()

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:
        self.__set_key_value(key, value)

    def __delitem__(self, key: str) -> None:
        if key in self.keys():
            self.__delattr__(key)
        else:
            raise IndexError("Key not found")

    def __set_key_value(self, key: str, value: Any, copy: bool = False) -> None:
        if value is None:
            value = {}

        match value:
            case dict():
                d = value.copy() if copy else value
                self.__setattr__(key, PfsSection(d))
            case PfsNonUniqueList():
                # multiple keywords/Sections with same name
                sections = PfsNonUniqueList()
                for v in value:
                    match v:
                        case dict():
                            d = v.copy() if copy else v
                            sections.append(PfsSection(d))
                        case _:
                            sections.append(self._parse_value(v))
                self.__setattr__(key, sections)
            case _:
                self.__setattr__(key, self._parse_value(value))

    def _parse_value(self, v: Any) -> Any:
        if isinstance(v, str) and self._str_is_scientific_float(v):
            return float(v)
        return v

    @staticmethod
    def _str_is_scientific_float(s: str) -> bool:
        """True: -1.0e2, 1E-4, -0.1E+0.5; False: E12, E-4."""
        if len(s) < 3 or s.lower().startswith('e'):
            return False
        try:
            float(s)
            return 'e' in s.lower()
        except ValueError:
            return False

    def pop(self, key: Any, default: Any = None) -> Any:
        """If key is in the dictionary, remove it and return its
        value, else return default. If default is not given and
        key is not in the dictionary, a KeyError is raised."""
        return self.__dict__.pop(key, default)

    def get(self, key: Any, default: Any = None) -> Any:
        """Return the value for key if key is in the PfsSection,
        else default. If default is not given, it defaults to None,
        so that this method never raises a KeyError."""
        return self.__dict__.get(key, default)

    def clear(self) -> None:
        """Remove all items from the PfsSection."""
        return self.__dict__.clear()

    def keys(self) -> KeysView[str]:
        """Return a new view of the PfsSection's keys."""
        return self.__dict__.keys()

    def values(self) -> ValuesView[Any]:
        """Return a new view of the PfsSection's values."""
        return self.__dict__.values()

    def items(self) -> ItemsView[str, Any]:
        """Return a new view of the PfsSection's items ((key, value) pairs)."""
        return self.__dict__.items()

    def search(
        self,
        text: str | None = None,
        *,
        key: str | None = None,
        section: str | None = None,
        param: str | bool | int | float | None = None,
        case: bool = False,
    ) -> PfsSection:
        """Find recursively all keys, sections or parameters
           matching a pattern.

        NOTE: logical OR between multiple conditions

        Parameters
        ----------
        text : str, optional
            Search for text in either key, section or parameter, by default None
        key : str, optional
            text pattern to seach for in keywords, by default None
        section : str, optional
            text pattern to seach for in sections, by default None
        param : str, bool, float, int, optional
            text or value in a parameter, by default None
        case : bool, optional
            should the text search be case-sensitive?, by default False

        Returns
        -------
        PfsSection
            Search result as a nested PfsSection

        """
        if text is not None:
            # text searches across all fields
            if key is not None or section is not None or param is not None:
                raise ValueError("When 'text' is provided, 'key', 'section' and 'param' must be None")
            key = section = param = text
            
        key = key.lower() if (key is not None and not case) else key
        section = section.lower() if (section is not None and not case) else section
        param = (
            param
            if (param is None or not isinstance(param, str) or case)
            else param.lower()
        )
        results = [item for item in self._find_patterns_generator(
            keypat=key, parampat=param, secpat=section, case=case
        )]
        return (
            self.__class__._merge_PfsSections(results)
            if len(results) > 0
            else PfsSection({})
        )

    def _find_patterns_generator(
        self,
        keypat: str | None = None,
        parampat: Any = None,
        secpat: str | None = None,
        keylist: list[str] | None = None,
        case: bool = False,
    ) -> Any:
        """Look for patterns in either keys, params or sections."""
        keylist = [] if keylist is None else keylist
        for k, v in self.items():
            kk = str(k) if case else str(k).lower()

            if isinstance(v, PfsSection):
                if secpat and secpat in kk:
                    yield from self._yield_deep_dict(keylist + [k], v)
                else:
                    yield from v._find_patterns_generator(
                        keypat, parampat, secpat, keylist=keylist + [k], case=case
                    )
            else:
                if (keypat and keypat in kk) or self._param_match(parampat, v, case):
                    yield from self._yield_deep_dict(keylist + [k], v)

    @staticmethod
    def _yield_deep_dict(keys: Sequence[str], val: Any) -> Any:
        """yield a deep nested dict with keys with a single deep value val."""
        for j in range(len(keys) - 1, -1, -1):
            d = {keys[j]: val}
            val = d
        yield d

    @staticmethod
    def _param_match(parampat: Any, v: Any, case: bool) -> Any:
        if parampat is None:
            return False
        if type(v) is not type(parampat):
            return False
        if isinstance(v, str):
            vv = str(v) if case else str(v).lower()
            return parampat in vv
        else:
            return parampat == v

    def find_replace(self, old_value: Any, new_value: Any) -> None:
        """Update recursively all old_value with new_value."""
        for k, v in self.items():
            if isinstance(v, PfsSection):
                self[k].find_replace(old_value, new_value)
            elif self[k] == old_value:
                self[k] = new_value

    def copy(self) -> "PfsSection":
        """Return a copy of the PfsSection."""
        return PfsSection(self.to_dict())

    def _to_txt_lines(self) -> list[str]:
        lines: list[str] = []
        self._append_to_lines_at_level(lines)
        return lines

    def _append_to_lines_at_level(self, lines: list[str], level: int = 0) -> None:
        lvl_prefix = "   " * level
        for k, v in self.items():
            # check for empty sections
            if v is None:
                lines.append(f"{lvl_prefix}[{k}]")
                lines.append(f"{lvl_prefix}EndSect  // {k}")

            elif isinstance(v, list) and any(
                isinstance(subv, PfsSection) for subv in v
            ):
                # duplicate sections
                for subv in v:
                    if isinstance(subv, PfsSection):
                        subsec = PfsSection({k: subv})
                        subsec._append_to_lines_at_level(lines, level=level)
                    else:
                        subv = self._prepare_value_for_write(subv)
                        lines.append(f"{lvl_prefix}{k} = {subv}")
            elif isinstance(v, PfsSection):
                lines.append(f"{lvl_prefix}[{k}]")
                v._append_to_lines_at_level(lines, level=(level + 1))
                lines.append(f"{lvl_prefix}EndSect  // {k}")
            elif isinstance(v, PfsNonUniqueList) or (
                isinstance(v, list) and all([isinstance(vv, list) for vv in v])
            ):
                if len(v) == 0:
                    # empty list -> keyword with no parameter
                    lines.append(f"{lvl_prefix}{k} = ")
                for subv in v:
                    psubv = self._prepare_value_for_write(subv)
                    lines.append(f"{lvl_prefix}{k} = {psubv}")
            else:
                pv = self._prepare_value_for_write(v)
                lines.append(f"{lvl_prefix}{k} = {pv}")

    def _prepare_value_for_write(
        self, v: str | bool | datetime | list[str | bool | datetime]
    ) -> str:
        """Catch peculiarities of string formatted pfs data."""
        match v:
            case str():
                if len(v) > 5 and not ("PROJ" in v or "<CLOB:" in v):
                    v = v.replace('"', "''")
                    v = v.replace("\U0001f600", "'")

                if v == "":
                    v = "''"
                elif v.count("|") == 2 and "CLOB" not in v:
                    v = f"{v}"
                else:
                    v = f"'{v}'"

            case bool():
                v = str(v).lower()  # stick to MIKE lowercase bool notation

            case datetime():
                v = v.strftime("%Y, %m, %d, %H, %M, %S").replace(" 0", " ")

            case list():
                out = []
                for subv in v:
                    out.append(str(self._prepare_value_for_write(subv)))
                v = ", ".join(out)

        return v

    def to_dict(self) -> dict[str, Any]:
        """Convert to (nested) dict (as a copy)."""
        d = self.__dict__.copy()
        for key, value in d.items():
            if isinstance(value, PfsSection):
                d[key] = value.to_dict()
        return d

    def to_dataframe(self, prefix: str | None = None) -> pd.DataFrame:
        """Output enumerated subsections to a DataFrame.

        Parameters
        ----------
        prefix : str, optional
            The prefix of the enumerated sections, e.g. "OUTPUT_",
            which can be supplied if it fails without this argument,
            by default None (will try to "guess" the prefix)

        Returns
        -------
        pd.DataFrame
            The enumerated subsections as a DataFrame

        Examples
        --------
        ```{python}
        pfs = mikeio.read_pfs("../data/pfs/lake.sw")
        pfs.SW.OUTPUTS.to_dataframe(prefix="OUTPUT_")
        ```

        """
        if prefix is not None:
            sections = [
                k for k in self.keys() if k.startswith(prefix) and k[-1].isdigit()
            ]
            n_sections = len(sections)
        else:
            n_sections = -1
            # TODO: check that value is a PfsSection
            sections = [k for k in self.keys() if k[-1].isdigit()]
            for k in self.keys():
                if isinstance(k, str) and k.startswith("number_of_"):
                    n_sections = self[k]
            if n_sections == -1:
                # raise ValueError("Could not find a number_of_... keyword")
                n_sections = len(sections)

        if len(sections) == 0:
            prefix_txt = "" if prefix is None else f"(starting with '{prefix}') "
            raise ValueError(f"No enumerated subsections {prefix_txt}found")

        prefix = sections[0][:-1]
        res = []
        for j in range(n_sections):
            k = f"{prefix}{j+1}"
            res.append(self[k].to_dict())
        return pd.DataFrame(res, index=range(1, n_sections + 1))

    @classmethod
    def _merge_PfsSections(cls, sections: Sequence[dict[str, Any]]) -> "PfsSection":
        """Merge a list of PfsSections/dict."""
        assert len(sections) > 0
        a = sections[0]
        for b in sections[1:]:
            a = _merge_dict(a, b)
        return cls(a)



from __future__ import annotations
from pathlib import Path
from typing import Any, Sequence, TYPE_CHECKING

from matplotlib.axes import Axes
import numpy as np
from mikecore.DfsuFile import DfsuFile, DfsuFileType
from mikecore.DfsFileFactory import DfsFileFactory
import pandas as pd
from scipy.spatial import cKDTree
from tqdm import trange

from ..dataset import DataArray, Dataset
from ..dfs._dfs import (
    _get_item_info,
    _read_item_time_step,
    _valid_item_numbers,
    _valid_timesteps,
)
from ..eum import EUMType, ItemInfo
from .._interpolation import get_idw_interpolant, interp2d
from ..spatial import (
    GeometryFM3D,
    GeometryFMVerticalProfile,
    GeometryPoint3D,
    GeometryFM2D,
)
from ..spatial._FM_utils import _plot_vertical_profile
from ._dfsu import (
    _get_dfsu_info,
    get_nodes_from_source,
    get_elements_from_source,
    _validate_elements_and_geometry_sel,
    write_dfsu_data,
)

if TYPE_CHECKING:
    from ..spatial._FM_geometry_layered import Layer


class DfsuLayered:
    show_progress = False

    def __init__(self, filename: str | Path) -> None:
        info = _get_dfsu_info(filename)
        self._filename = info.filename
        self._type = info.type
        self._deletevalue = info.deletevalue
        self._equidistant = info.equidistant
        self._start_time = info.start_time
        self._timestep = info.timestep
        self._n_timesteps = info.n_timesteps
        self._geometry = self._read_geometry(self._filename)
        # 3d files have a zn item
        self._items = self._read_items(self._filename)

    def __repr__(self) -> str:
        out = [
            f"<mikeio.{self.__class__.__name__}>"
            f"number of nodes: {self.geometry.n_nodes}",
            f"number of elements: {self.geometry.n_elements}",
            f"projection: {self.geometry.projection_string}",
            f"number of sigma layers: {self.geometry.n_sigma_layers}",
        ]
        if (
            self._type == DfsuFileType.DfsuVerticalProfileSigmaZ
            or self._type == DfsuFileType.Dfsu3DSigmaZ
        ):
            out.append(
                f"max number of z layers: {self.geometry.n_layers - self.geometry.n_sigma_layers}"
            )
        if self.n_items < 10:
            out.append("items:")
            for i, item in enumerate(self.items):
                out.append(f"  {i}:  {item}")
        else:
            out.append(f"number of items: {self.n_items}")
        if self.n_timesteps == 1:
            out.append(f"time: time-invariant file (1 step) at {self.time[0]}")
        else:
            out.append(
                f"time: {str(self.time[0])} - {str(self.time[-1])} ({self.n_timesteps} records)"
            )
        return str.join("\n", out)

    @property
    def deletevalue(self) -> float:
        """File delete value."""
        return self._deletevalue

    @property
    def n_items(self) -> int:
        """Number of items."""
        return len(self.items)

    @property
    def items(self) -> list[ItemInfo]:
        """List of items."""
        return self._items

    @property
    def start_time(self) -> pd.Timestamp:
        """File start time."""
        return self._start_time

    @property
    def n_timesteps(self) -> int:
        """Number of time steps."""
        return self._n_timesteps

    @property
    def timestep(self) -> float:
        """Time step size in seconds."""
        return self._timestep

    @property
    def end_time(self) -> pd.Timestamp:
        """File end time."""
        if self._equidistant:
            return self.time[-1]
        else:
            # read the last timestep
            ds = self.read(items=0, time=-1)
            return ds.time[-1]

    @property
    def time(self) -> pd.DatetimeIndex:
        if self._equidistant:
            return pd.date_range(
                start=self.start_time,
                periods=self.n_timesteps,
                freq=f"{int(self.timestep)}s",
            )
        else:
            raise NotImplementedError(
                "Non-equidistant time axis. Read the data to get time."
            )

    @property
    def geometry(self) -> GeometryFM3D | GeometryFMVerticalProfile:
        return self._geometry

    @staticmethod
    def _read_items(filename: str) -> list[ItemInfo]:
        dfs = DfsuFile.Open(filename)
        n_items = len(dfs.ItemInfo)
        first_idx = 1
        items = _get_item_info(
            dfs.ItemInfo,
            list(range(n_items - first_idx)),
            ignore_first=True,
        )
        dfs.Close()
        return items

    @staticmethod
    def _read_geometry(filename: str) -> GeometryFM3D | GeometryFMVerticalProfile:
        dfs = DfsuFile.Open(filename)
        dfsu_type = DfsuFileType(dfs.DfsuFileType)

        node_table = get_nodes_from_source(dfs)
        el_table = get_elements_from_source(dfs)

        geom_cls: Any = GeometryFM3D
        if dfsu_type in (
            DfsuFileType.DfsuVerticalProfileSigma,
            DfsuFileType.DfsuVerticalProfileSigmaZ,
        ):
            geom_cls = GeometryFMVerticalProfile

        geometry = geom_cls(
            node_coordinates=node_table.coordinates,
            element_table=el_table.connectivity,
            codes=node_table.codes,
            projection=dfs.Projection.WKTString,
            dfsu_type=dfsu_type,
            element_ids=el_table.ids,
            node_ids=node_table.ids,
            n_layers=dfs.NumberOfLayers,
            n_sigma=min(dfs.NumberOfSigmaLayers, dfs.NumberOfLayers),
            validate=False,
        )
        dfs.Close()
        return geometry

    @property
    def n_layers(self) -> int:
        """Maximum number of layers."""
        return self.geometry._n_layers

    @property
    def n_sigma_layers(self) -> int:
        """Number of sigma layers."""
        return self.geometry.n_sigma_layers

    @property
    def n_z_layers(self) -> int:
        """Maximum number of z-layers."""
        return self.n_layers - self.n_sigma_layers

    def read(
        self,
        *,
        items: str | int | Sequence[str | int] | None = None,
        time: int | str | slice | None = None,
        elements: Sequence[int] | np.ndarray | None = None,
        area: tuple[float, float, float, float] | None = None,
        x: float | None = None,
        y: float | None = None,
        z: float | None = None,
        layers: int | Layer | Sequence[int] | None = None,
        keepdims: bool = False,
        dtype: Any = np.float32,
        error_bad_data: bool = True,
        fill_bad_data_value: float = np.nan,
    ) -> Dataset:
        """Read data from a dfsu file.

        Parameters
        ---------
        items: list[int] or list[str], optional
            Read only selected items, by number (0-based), or by name
        time: int, str, datetime, pd.TimeStamp, sequence, slice or pd.DatetimeIndex, optional
            Read only selected time steps, by default None (=all)
        keepdims: bool, optional
            When reading a single time step only, should the time-dimension be kept
            in the returned Dataset? by default: False
        area: list[float], optional
            Read only data inside (horizontal) area given as a
            bounding box (tuple with left, lower, right, upper)
            or as list of coordinates for a polygon, by default None
        x: float, optional
            Read only data for elements containing the (x,y,z) points(s)
        y: float, optional
            Read only data for elements containing the (x,y,z) points(s)
        z: float, optional
            Read only data for elements containing the (x,y,z) points(s)
        layers: int, str, list[int], optional
            Read only data for specific layers, by default None
        elements: list[int], optional
            Read only selected element ids, by default None
        error_bad_data: bool, optional
            raise error if data is corrupt, by default True,
        fill_bad_data_value:
            fill value for to impute corrupt data, used in conjunction with error_bad_data=False
            default np.nan
        dtype: numpy.dtype, optional
            Data type to read, by default np.float32

        Returns
        -------
        Dataset
            A Dataset with data dimensions [t,elements]

        """
        if dtype not in [np.float32, np.float64]:
            raise ValueError("Invalid data type. Choose np.float32 or np.float64")

        dfs = DfsuFile.Open(self._filename)

        single_time_selected, time_steps = _valid_timesteps(dfs, time)

        _validate_elements_and_geometry_sel(
            elements, area=area, layers=layers, x=x, y=y, z=z
        )
        if elements is None:
            if (
                (x is not None)
                or (y is not None)
                or (area is not None)
                or (layers is not None)
            ):
                elements = self.geometry.find_index(  # type: ignore
                    x=x, y=y, z=z, area=area, layers=layers
                )
                if len(elements) == 0:  # type: ignore
                    raise ValueError("No elements in selection!")

        geometry = (
            self.geometry
            if elements is None
            else self.geometry.elements_to_geometry(elements)
        )

        if isinstance(geometry, GeometryPoint3D):
            n_elems = 1
            n_nodes = 1
        else:
            n_elems = geometry.n_elements
            n_nodes = geometry.n_nodes
            node_ids = geometry.node_ids

        item_numbers = _valid_item_numbers(
            dfs.ItemInfo, items, ignore_first=self.geometry.is_layered
        )
        items = _get_item_info(
            dfs.ItemInfo, item_numbers, ignore_first=self.geometry.is_layered
        )
        item_numbers = [it + 1 for it in item_numbers]
        if layered_data := hasattr(geometry, "is_layered") and geometry.is_layered:
            item_numbers.insert(0, 0)
        n_items = len(item_numbers)

        deletevalue = self.deletevalue

        data_list = []
        t_seconds = np.zeros(len(time_steps))
        n_steps = len(time_steps)
        item0_is_node_based = False
        for item in range(n_items):
            if layered_data and item == 0:
                item0_is_node_based = True
                data: np.ndarray = np.ndarray(shape=(n_steps, n_nodes), dtype=dtype)
            else:
                data = np.ndarray(shape=(n_steps, n_elems), dtype=dtype)
            data_list.append(data)

        if single_time_selected and not keepdims:
            data = data[0]

        for i in trange(n_steps, disable=not self.show_progress):
            it = time_steps[i]
            for item in range(n_items):
                dfs, d, t = _read_item_time_step(
                    dfs=dfs,
                    filename=self._filename,
                    time=time,
                    item_numbers=item_numbers,
                    deletevalue=deletevalue,
                    shape=(data.shape[-1],),
                    item=item,
                    it=it,
                    error_bad_data=error_bad_data,
                    fill_bad_data_value=fill_bad_data_value,
                )
                t_seconds[i] = t

                if elements is not None:
                    if item == 0 and item0_is_node_based:
                        d = d[node_ids]
                    else:
                        d = d[elements]  # type: ignore

                if single_time_selected and not keepdims:
                    data_list[item] = d
                else:
                    data_list[item][i] = d

        time = pd.to_datetime(t_seconds, unit="s", origin=self.start_time)

        dfs.Close()

        dims = (
            ("time", "element")
            if not (single_time_selected and not keepdims)  # TODO extract variable
            else ("element",)
        )

        if elements is not None and len(elements) == 1:
            # squeeze point data
            dims = tuple([d for d in dims if d != "element"])
            data_list = [np.squeeze(d, axis=-1) for d in data_list]

        if layered_data:
            return Dataset(
                data_list[1:],  # skip zn item
                time,
                items,
                geometry=geometry,
                zn=data_list[0],
                dims=dims,
                validate=False,
                dt=self.timestep,
            )
        else:
            return Dataset(
                data_list,
                time,
                items,
                geometry=geometry,
                dims=dims,
                validate=False,
                dt=self.timestep,
            )

    def append(self, ds: Dataset, validate: bool = True) -> None:
        """Append data to a dfsu file.

        Parameters
        ---------
        ds: Dataset
            Dataset to append
        validate: bool, optional
            Validate that the dataset to append has the same geometry and items, by default True

        """
        if validate:
            if self.geometry != ds.geometry:
                raise ValueError("The geometry of the dataset to append does not match")

            for item_s, item_o in zip(ds.items, self.items):
                if item_s != item_o:
                    raise ValueError(
                        f"Item in dataset {item_s.name} does not match {item_o.name}"
                    )

        dfs = DfsFileFactory.DfsuFileOpenAppend(str(self._filename), parameters=None)
        write_dfsu_data(dfs=dfs, ds=ds, is_layered=ds.geometry.is_layered)
        info = _get_dfsu_info(self._filename)
        self._n_timesteps = info.n_timesteps


class Dfsu2DV(DfsuLayered):
    """Class for reading/writing dfsu 2d vertical files."""

    def plot_vertical_profile(
        self,
        values: np.ndarray | DataArray,
        cmin: float | None = None,
        cmax: float | None = None,
        label: str = "",
        **kwargs: Any,
    ) -> Axes:
        """Plot unstructured vertical profile.

        Parameters
        ----------
        values: np.array
            value for each element to plot
        cmin: real, optional
            lower bound of values to be shown on plot, default:None
        cmax: real, optional
            upper bound of values to be shown on plot, default:None
        title: str, optional
            axes title
        label: str, optional
            colorbar label
        cmap: matplotlib.cm.cmap, optional
            colormap, default viridis
        figsize: (float, float), optional
            specify size of figure
        ax: matplotlib.axes, optional
            Adding to existing axis, instead of creating new fig
        **kwargs: Any
            Additional keyword arguments

        Returns
        -------
        <matplotlib.axes>

        """
        if isinstance(values, DataArray):
            values = values.to_numpy()

        g = self.geometry
        return _plot_vertical_profile(
            node_coordinates=g.node_coordinates,
            element_table=g.element_table,
            values=values,
            zn=None,
            is_geo=g.is_geo,
            cmin=cmin,
            cmax=cmax,
            label=label,
            **kwargs,
        )


class Dfsu3D(DfsuLayered):
    """Class for reading/writing dfsu 3d files."""

    @property
    def geometry2d(self) -> GeometryFM2D:
        """The 2d geometry for a 3d object."""
        return self.geometry.geometry2d

    def extract_surface_elevation_from_3d(self, n_nearest: int = 4) -> DataArray:
        """Extract surface elevation from a 3d dfsu file (based on zn)
        to a new 2d dfsu file with a surface elevation item.

        Parameters
        ---------
        n_nearest: int, optional
            number of points for spatial interpolation (inverse_distance), default=4

        """
        # validate input
        assert (
            self._type == DfsuFileType.Dfsu3DSigma
            or self._type == DfsuFileType.Dfsu3DSigmaZ
        )
        assert n_nearest > 0

        # make 2d nodes-to-elements interpolator
        top_el = self.geometry.top_elements
        geom = self.geometry.elements_to_geometry(top_el, node_layers="top")
        xye = geom.element_coordinates[:, 0:2]  # type: ignore
        xyn = geom.node_coordinates[:, 0:2]  # type: ignore
        tree2d = cKDTree(xyn)
        dist, node_ids = tree2d.query(xye, k=n_nearest)
        if n_nearest == 1:
            weights = None
        else:
            weights = get_idw_interpolant(dist)

        # read zn from 3d file and interpolate to element centers
        ds = self.read(items=0, keepdims=True)  # read only zn
        node_ids_surf, _ = self.geometry._get_nodes_and_table_for_elements(
            top_el, node_layers="top"
        )
        assert isinstance(ds[0]._zn, np.ndarray)
        zn_surf = ds[0]._zn[:, node_ids_surf]  # surface
        surf2d = interp2d(zn_surf, node_ids, weights)
        surf_da = DataArray(
            data=surf2d,
            time=ds.time,
            geometry=geom,
            item=ItemInfo(EUMType.Surface_Elevation),
        )

        return surf_da



from __future__ import annotations
from pathlib import Path
import warnings
from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Sequence
import numpy as np
import pandas as pd

from mikecore.DfsFile import (
    DfsDynamicItemInfo,
    DfsFile,
    DfsFileInfo,
    TimeAxisType,
)
from mikecore.DfsFileFactory import DfsFileFactory
from mikecore.Projections import Cartography

from ..dataset import Dataset
from ..eum import ItemInfo, ItemInfoList
from ..exceptions import ItemsError
from .._time import DateTimeSelector


@dataclass
class DfsHeader:
    n_items: int
    n_timesteps: int
    start_time: datetime
    dt: float
    coordinates: tuple[str, float, float, float]
    items: list[ItemInfo]


def _read_item_time_step(
    *,
    dfs: DfsFile,
    filename: str,
    time: pd.DatetimeIndex,
    item_numbers: list[int],
    deletevalue: float,
    shape: tuple[int, ...],
    item: int,
    it: int,
    error_bad_data: bool = True,
    fill_bad_data_value: float = np.nan,
) -> tuple[DfsFile, np.ndarray, float]:
    itemdata = dfs.ReadItemTimeStep(item_numbers[item] + 1, it)
    t = itemdata.Time
    if itemdata is not None:
        d = itemdata.Data
        d[d == deletevalue] = np.nan
    else:
        if error_bad_data:
            raise ValueError(f"Error reading: {time[it]}")
        else:
            warnings.warn(f"Error reading: {time[it]}")
            d = np.zeros(shape[1])
            d[:] = fill_bad_data_value
            dfs.Close()
            dfs = DfsFileFactory.DfsGenericOpen(filename)
    return dfs, d, t


def _fuzzy_item_search(
    *, dfsItemInfo: list[DfsDynamicItemInfo], search: str, start_idx: int = 0
) -> list[int]:
    import fnmatch

    names = [info.Name for info in dfsItemInfo]
    item_numbers = [
        i - start_idx for i, name in enumerate(names) if fnmatch.fnmatch(name, search)
    ]
    if len(item_numbers) == 0:
        raise KeyError(f"No items like: {search} found. Valid names are {names}")
    return item_numbers


def _valid_item_numbers(
    dfsItemInfo: list[DfsDynamicItemInfo],
    items: str | int | Sequence[int | str] | None = None,
    ignore_first: bool = False,
) -> list[int]:
    start_idx = 1 if ignore_first else 0
    n_items_file = len(dfsItemInfo) - start_idx
    if items is None:
        return list(range(n_items_file))

    # Handling scalar and sequences is a bit tricky

    item_numbers: list[int] = []

    # check if items is a scalar (int or str)
    if isinstance(items, (int, str)):
        if isinstance(items, str) and "*" in items:
            return _fuzzy_item_search(
                dfsItemInfo=dfsItemInfo, search=items, start_idx=start_idx
            )
        elif isinstance(items, str):
            item_number = _item_numbers_by_name(dfsItemInfo, [items], ignore_first)[0]
            return [item_number]
        elif isinstance(items, int):
            if (items < 0) or (items >= n_items_file):
                raise ItemsError(n_items_file)
            return [items]

    assert isinstance(items, Sequence)
    for item in items:
        if isinstance(item, str):
            item_number = _item_numbers_by_name(dfsItemInfo, [item], ignore_first)[0]
        elif isinstance(item, int):
            if (item < 0) or (item >= n_items_file):
                raise ItemsError(n_items_file)
            item_number = item
        else:
            raise ItemsError(n_items_file)
        item_numbers.append(item_number)

    if len(set(item_numbers)) != len(item_numbers):
        raise ValueError("'items' must be unique")

    return item_numbers


def _valid_timesteps(
    dfsFileInfo: DfsFileInfo, time_steps: int | Sequence[int] | str | slice | None
) -> tuple[bool, list[int]]:
    time_axis = dfsFileInfo.TimeAxis

    single_time_selected = False
    if isinstance(time_steps, (int, datetime)):
        single_time_selected = True

    nt = time_axis.NumberOfTimeSteps

    if time_axis.TimeAxisType != TimeAxisType.CalendarEquidistant:
        # TODO is this the proper epoch, should this magic number be somewhere else?
        start_time_file = datetime(1970, 1, 1)
    else:
        start_time_file = time_axis.StartDateTime

    if time_axis.TimeAxisType in (
        TimeAxisType.CalendarEquidistant,
        TimeAxisType.TimeEquidistant,
    ):
        time_step_file = time_axis.TimeStep

        if time_step_file <= 0:
            if nt > 1:
                raise ValueError(
                    f"Time step must be a positive number. Time step in the file is {time_step_file} seconds."
                )

            warnings.warn(
                f"Time step is {time_step_file} seconds. This must be a positive number. Setting to 1 second."
            )
            time_step_file = 1

        freq = pd.Timedelta(seconds=time_step_file)
        time = pd.date_range(start_time_file, periods=nt, freq=freq)
    elif time_axis.TimeAxisType == TimeAxisType.CalendarNonEquidistant:
        idx = list(range(nt))

        if isinstance(time_steps, int):
            return True, [idx[time_steps]]
        return single_time_selected, idx

    dts = DateTimeSelector(time)

    idx = dts.isel(time_steps)

    if isinstance(time_steps, str):
        if len(idx) == 1:
            single_time_selected = True

    return single_time_selected, idx


def _item_numbers_by_name(
    dfsItemInfo: list[DfsDynamicItemInfo],
    item_names: list[str],
    ignore_first: bool = False,
) -> list[int]:
    """Utility function to find item numbers.

    Parameters
    ----------
    dfsItemInfo : list[DfsDynamicItemInfo]
        item info from dfs file
    item_names : list[str]
        Names of items to be found
    ignore_first : bool, optional
        Ignore first item, by default False


    Returns
    -------
    list[int]
        item numbers (0-based)

    Raises
    ------
    KeyError
        In case item is not found in the dfs file

    """
    first_idx = 1 if ignore_first else 0
    names = [x.Name for x in dfsItemInfo[first_idx:]]

    item_lookup = {name: i for i, name in enumerate(names)}
    try:
        item_numbers = [item_lookup[x] for x in item_names]
    except KeyError:
        raise KeyError(f"Selected item name not found. Valid names are {names}")

    return item_numbers


def _get_item_info(
    dfsItemInfo: list[DfsDynamicItemInfo],
    item_numbers: list[int] | None = None,
    ignore_first: bool = False,
) -> ItemInfoList:
    """Read DFS ItemInfo for specific item numbers.

    Parameters
    ----------
    dfsItemInfo : list[DfsDynamicItemInfo]
        Item info from dfs file
    item_numbers : list[int], optional
        Item numbers to read, by default all items are read
    ignore_first : bool, optional
        Ignore first item, by default False, used for Dfsu3D

    Returns
    -------
    ItemInfoList

    """
    first_idx = 1 if ignore_first else 0
    if item_numbers is None:
        item_numbers = list(range(len(dfsItemInfo) - first_idx))

    item_numbers = [i + first_idx for i in item_numbers]
    items = [
        ItemInfo.from_mikecore_dynamic_item_info(dfsItemInfo[i]) for i in item_numbers
    ]
    return ItemInfoList(items)


def write_dfs_data(*, dfs: DfsFile, ds: Dataset, n_spatial_dims: int) -> None:
    deletevalue = dfs.FileInfo.DeleteValueFloat  # ds.deletevalue
    has_no_time = "time" not in ds.dims
    if ds.is_equidistant:
        t_rel = np.zeros(ds.n_timesteps)
    else:
        t_rel = (ds.time - ds.time[0]).total_seconds()

    for i in range(ds.n_timesteps):
        for item in range(ds.n_items):
            if has_no_time:
                d = ds[item].values
            else:
                d = ds[item].values[i]
            d = d.copy()  # to avoid modifying the input
            d[np.isnan(d)] = deletevalue

            d = d.reshape(ds.shape[-n_spatial_dims:])  # spatial axes
            darray = d.flatten()

            dfs.WriteItemTimeStepNext(t_rel[i], darray.astype(np.float32))

    dfs.Close()


class _Dfs123:
    _ndim: int

    show_progress = False

    def __init__(self, filename: str | Path) -> None:
        path = Path(filename)
        if not path.exists():
            raise FileNotFoundError(path)
        self._filename = str(filename) if filename else None
        self._end_time = None
        self._is_equidistant = True
        dfs = DfsFileFactory.DfsGenericOpen(self._filename)
        self._dfs = dfs
        self._geometry: Any = None  # Handled by subclass
        n_items = len(dfs.ItemInfo)
        self._items = self._get_item_info(list(range(n_items)))
        if dfs.FileInfo.TimeAxis.TimeAxisType in {
            TimeAxisType.CalendarEquidistant,
            TimeAxisType.CalendarNonEquidistant,
        }:
            self._start_time = dfs.FileInfo.TimeAxis.StartDateTime
        else:  # relative time axis
            self._start_time = datetime(
                1970, 1, 1
            )  # TODO is this the proper epoch, should this magic number be somewhere else?

        if hasattr(dfs.FileInfo.TimeAxis, "TimeStep"):
            self._timestep = (
                # some files have dt = 0 😳
                dfs.FileInfo.TimeAxis.TimeStepInSeconds()
                if dfs.FileInfo.TimeAxis.TimeStepInSeconds() > 0
                else 1
            )  # TODO handle other timeunits

            freq = pd.Timedelta(seconds=self._timestep)
            self._time = pd.date_range(
                start=self._start_time,
                periods=dfs.FileInfo.TimeAxis.NumberOfTimeSteps,
                freq=freq,
            )
        else:
            self._timestep = None
            self._time = None
        self._n_timesteps: int = dfs.FileInfo.TimeAxis.NumberOfTimeSteps
        projstr = dfs.FileInfo.Projection.WKTString
        self._projstr: str = "NON-UTM" if not projstr else projstr
        self._longitude: float = dfs.FileInfo.Projection.Longitude
        self._latitude: float = dfs.FileInfo.Projection.Latitude
        self._orientation: float = dfs.FileInfo.Projection.Orientation
        self._deletevalue: float = dfs.FileInfo.DeleteValueFloat

        dfs.Close()

    def __repr__(self) -> str:
        name = self.__class__.__name__
        out = [f"<mikeio.{name}>"]

        out.append(f"geometry: {self.geometry}")
        if self.n_items < 10:
            out.append("items:")
            for i, item in enumerate(self.items):
                out.append(f"  {i}:  {item}")
        else:
            out.append(f"number of items: {self.n_items}")

        if self._n_timesteps == 1:
            out.append("time: time-invariant file (1 step)")
        else:
            out.append(f"time: {self._n_timesteps} steps")
            out.append(f"start time: {self._start_time}")

        return str.join("\n", out)

    def _open(self) -> None:
        raise NotImplementedError("Should be implemented by subclass")

    def _get_item_info(self, item_numbers: Sequence[int]) -> list[ItemInfo]:
        """Read DFS ItemInfo.

        Parameters
        ----------
        item_numbers : list[int]
            Item numbers to read

        Returns
        -------
        list[Iteminfo]

        """
        infos = self._dfs.ItemInfo
        nos = item_numbers
        return [ItemInfo.from_mikecore_dynamic_item_info(infos[i]) for i in nos]

    @property
    def geometry(self) -> Any:
        return self._geometry

    @property
    def deletevalue(self) -> float:
        "File delete value."
        return self._deletevalue

    @property
    def n_items(self) -> int:
        "Number of items."
        return len(self.items)

    @property
    def items(self) -> list[ItemInfo]:
        "List of items."
        return self._items

    @property
    def time(self) -> pd.DatetimeIndex | None:
        return self._time

    @property
    def start_time(self) -> pd.Timestamp:
        """File start time."""
        return self._start_time

    @property
    def end_time(self) -> pd.Timestamp:
        """File end time."""
        if self._end_time is None:
            self._end_time = self.read(items=[0]).time[-1].to_pydatetime()  # type: ignore

        return self._end_time

    @property
    def n_timesteps(self) -> int:
        """Number of time steps."""
        return self._n_timesteps

    @property
    def timestep(self) -> Any:
        """Time step size in seconds."""
        # this will fail if the TimeAxisType is not calendar and equidistant, but that is ok
        return self._dfs.FileInfo.TimeAxis.TimeStepInSeconds()

    @property
    def projection_string(self) -> str:
        return self._projstr

    @property
    def longitude(self) -> float:
        """Origin longitude."""
        return self._longitude

    @property
    def latitude(self) -> float:
        """Origin latitude."""
        return self._latitude

    @property
    def origin(self) -> Any:
        """Origin (in own CRS)."""
        return self.geometry.origin

    @property
    def orientation(self) -> Any:
        """Orientation (in own CRS)."""
        return self.geometry.orientation

    @property
    def is_geo(self) -> bool:
        """Are coordinates geographical (LONG/LAT)?"""
        return self._projstr == "LONG/LAT"

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        """Shape of the data array."""
        pass

    def _validate_no_orientation_in_geo(self) -> None:
        if self.is_geo and abs(self._orientation) > 1e-6:
            raise ValueError("Orientation is not supported for LONG/LAT coordinates")

    def _origin_and_orientation_in_CRS(self) -> tuple[Any, float]:
        """Project origin and orientation to projected CRS (if not LONG/LAT)."""
        if self.is_geo:
            origin = self._longitude, self._latitude
            orientation = 0.0
        else:
            lon, lat = self._longitude, self._latitude
            cart = Cartography.CreateGeoOrigin(
                projectionString=self._projstr,
                lonOrigin=lon,
                latOrigin=lat,
                orientation=self._orientation,
            )
            # convert origin and orientation to projected CRS
            origin = tuple(np.round(cart.Geo2Proj(lon, lat), 6))  # type: ignore
            orientation = cart.OrientationProj
        return origin, orientation



from __future__ import annotations

import numpy as np

from ._geometry import BoundingBox


def xy_to_bbox(xy: np.ndarray, buffer: float = 0.0) -> BoundingBox:
    """Return bounding box for list of coordinates."""
    left = xy[:, 0].min() - buffer
    bottom = xy[:, 1].min() - buffer
    right = xy[:, 0].max() + buffer
    top = xy[:, 1].max() + buffer
    return BoundingBox(left, bottom, right, top)


def dist_in_meters(
    coords: np.ndarray, pt: tuple[float, float], is_geo: bool = False
) -> np.ndarray:
    """Get distance between array of coordinates and point.

    Parameters
    ----------
    coords : n-by-2 array
        x, y coordinates
    pt : [float, float]
        x, y coordinate of point
    is_geo : bool, optional
        are coordinates geographical?, by default False

    Returns
    -------
    array
        distances in meter

    """
    coords = np.atleast_2d(coords)
    xe = coords[:, 0]
    ye = coords[:, 1]
    xp = pt[0]
    yp = pt[1]
    if is_geo:
        d = _get_dist_geo(xe, ye, xp, yp)
    else:
        d = np.sqrt(np.square(xe - xp) + np.square(ye - yp))
    return d  # type: ignore


def _get_dist_geo(lon: float, lat: float, lon1: float, lat1: float) -> float:
    # assuming input in degrees!
    R = 6371e3  # Earth radius in metres
    dlon = np.deg2rad(lon1 - lon)
    dlon[dlon > np.pi] = dlon[dlon > np.pi] - 2 * np.pi
    dlon[dlon < -np.pi] = dlon[dlon < -np.pi] + 2 * np.pi
    dlat = np.deg2rad(lat1 - lat)
    x = dlon * np.cos(np.deg2rad((lat + lat1) / 2))
    y = dlat
    d = R * np.sqrt(np.square(x) + np.square(y))
    return d  # type: ignore


def _relative_cumulative_distance(
    coords: np.ndarray, reference: np.ndarray | None = None, is_geo: bool = False
) -> np.ndarray:
    """Calculate the cumulative relative distance along a path."""
    coords = np.atleast_2d(coords)
    d = np.zeros_like(coords[:, 0])
    if reference is not None:
        pt = (reference[0], reference[1])
        d[0] = dist_in_meters(coords[0, 0:2], pt=pt, is_geo=is_geo)[0]
    for j in range(1, len(d)):
        d[j] = d[j - 1] + dist_in_meters(coords[j, 0:2], coords[j - 1, 0:2], is_geo)[0]
    return d  # type: ignore



from ._geometry import GeometryPoint3D, GeometryPoint2D, GeometryUndefined
from ._FM_geometry import (
    GeometryFM2D,
)

from ._FM_geometry_spectral import (
    GeometryFMPointSpectrum,
    GeometryFMLineSpectrum,
    GeometryFMAreaSpectrum,
)

from ._FM_geometry_layered import (
    GeometryFMVerticalColumn,
    GeometryFMVerticalProfile,
    GeometryFM3D,
)

from ._grid_geometry import Grid1D, Grid2D, Grid3D


__all__ = [
    "GeometryPoint3D",
    "GeometryPoint2D",
    "GeometryUndefined",
    "GeometryFM2D",
    "GeometryFM3D",
    "GeometryFMPointSpectrum",
    "GeometryFMLineSpectrum",
    "GeometryFMAreaSpectrum",
    "GeometryFMVerticalColumn",
    "GeometryFMVerticalProfile",
    "Grid1D",
    "Grid2D",
    "Grid3D",
]



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING


import numpy as np


from mikecore.MeshFile import MeshFile


from ..spatial import GeometryFM2D
from ._common import (
    get_elements_from_source,
    get_nodes_from_source,
)

if TYPE_CHECKING:
    from shapely.geometry import MultiPolygon


class Mesh:
    """The Mesh class is initialized with a mesh file.

    Parameters
    ---------
    filename: str
        mesh filename

    Attributes
    ----------
    geometry: GeometryFM2D
        Flexible Mesh geometry

    Examples
    --------
    ```{python}
    import mikeio
    mikeio.Mesh("../data/odense_rough.mesh")
    ```

    """

    def __init__(self, filename: str | Path) -> None:
        self.geometry: GeometryFM2D = self._read_header(filename)
        self.plot = self.geometry.plot

    def _read_header(self, filename: str | Path) -> GeometryFM2D:
        msh = MeshFile.ReadMesh(filename)

        node_table = get_nodes_from_source(msh)
        el_table = get_elements_from_source(msh)

        geom = GeometryFM2D(
            node_coordinates=node_table.coordinates,
            element_table=el_table.connectivity,
            codes=node_table.codes,
            projection=msh.ProjectionString,
            element_ids=el_table.ids,
            node_ids=node_table.ids,
            validate=False,
        )

        return geom

    def __repr__(self) -> str:
        out = [
            "<Mesh>",
            f"number of nodes: {self.n_nodes}",
            f"number of elements: {self.n_elements}",
            f"projection: {self.geometry.projection_string}",
        ]
        return str.join("\n", out)

    # TODO re-consider if all of these properties are needed, since they all are available in the geometry
    @property
    def n_elements(self) -> int:
        """Number of elements."""
        return self.geometry.n_elements

    @property
    def element_coordinates(self) -> np.ndarray:
        """Coordinates of element centroids."""
        return self.geometry.element_coordinates

    @property
    def node_coordinates(self) -> np.ndarray:
        """Coordinates of nodes."""
        return self.geometry.node_coordinates

    @property
    def n_nodes(self) -> int:
        """Number of nodes."""
        return self.geometry.n_nodes

    @property
    def codes(self) -> np.ndarray:
        """Codes of nodes."""
        return self.geometry.codes

    @property
    def element_table(self) -> np.ndarray:
        """Element table."""
        return self.geometry.element_table

    @property
    def zn(self) -> np.ndarray:
        """Static bathymetry values (depth) at nodes."""
        return self.geometry.node_coordinates[:, 2]

    @zn.setter
    def zn(self, v: np.ndarray) -> None:
        if len(v) != self.n_nodes:
            raise ValueError(f"zn must have length of nodes ({self.n_nodes})")
        self.geometry.node_coordinates[:, 2] = v

    def write(
        self,
        outfilename: str | Path,
    ) -> None:
        """write mesh to file.

        Parameters
        ----------
        outfilename : str
            path to file

        """
        geometry = self.geometry

        assert isinstance(geometry, GeometryFM2D)  # i.e. not a GeometryPoint2d

        self.geometry.to_mesh(outfilename=outfilename)

    def to_shapely(self) -> MultiPolygon:
        """Convert Mesh geometry to shapely MultiPolygon.

        Returns
        -------
        MultiPolygon
            mesh as shapely MultiPolygon

        Examples
        --------
        ```{python}
        import mikeio
        msh = mikeio.open("../data/odense_rough.mesh")
        msh.to_shapely()
        ```

        """
        return self.geometry.to_shapely()
